#
# Common code for all guests
#
# Copyright 2006-2009, 2013, 2014, 2015 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import random

from . import generatename
from . import xmlutil
from .buildconfig import BuildConfig
from .devices import *  # pylint: disable=wildcard-import
from .domain import *  # pylint: disable=wildcard-import
from .domcapabilities import DomainCapabilities
from .logger import log
from .osdict import OSDB
from .xmlbuilder import XMLBuilder, XMLProperty, XMLChildProperty

_ignore = Device


class _DomainDevices(XMLBuilder):
    XML_NAME = "devices"
    _XML_PROP_ORDER = [
        "disk",
        "controller",
        "filesystem",
        "interface",
        "smartcard",
        "serial",
        "parallel",
        "console",
        "channel",
        "input",
        "tpm",
        "graphics",
        "sound",
        "audio",
        "video",
        "hostdev",
        "redirdev",
        "watchdog",
        "memballoon",
        "rng",
        "panic",
        "shmem",
        "memory",
        "vsock",
        "iommu",
        "pstore",
    ]

    disk = XMLChildProperty(DeviceDisk)
    controller = XMLChildProperty(DeviceController)
    filesystem = XMLChildProperty(DeviceFilesystem)
    interface = XMLChildProperty(DeviceInterface)
    smartcard = XMLChildProperty(DeviceSmartcard)
    serial = XMLChildProperty(DeviceSerial)
    parallel = XMLChildProperty(DeviceParallel)
    console = XMLChildProperty(DeviceConsole)
    channel = XMLChildProperty(DeviceChannel)
    input = XMLChildProperty(DeviceInput)
    tpm = XMLChildProperty(DeviceTpm)
    graphics = XMLChildProperty(DeviceGraphics)
    sound = XMLChildProperty(DeviceSound)
    audio = XMLChildProperty(DeviceAudio)
    video = XMLChildProperty(DeviceVideo)
    hostdev = XMLChildProperty(DeviceHostdev)
    redirdev = XMLChildProperty(DeviceRedirdev)
    watchdog = XMLChildProperty(DeviceWatchdog)
    memballoon = XMLChildProperty(DeviceMemballoon)
    rng = XMLChildProperty(DeviceRng)
    panic = XMLChildProperty(DevicePanic)
    shmem = XMLChildProperty(DeviceShMem)
    memory = XMLChildProperty(DeviceMemory)
    vsock = XMLChildProperty(DeviceVsock)
    iommu = XMLChildProperty(DeviceIommu)
    pstore = XMLChildProperty(DevicePstore)

    def get_all(self):
        retlist = []
        # pylint: disable=protected-access
        devtypes = _DomainDevices._XML_PROP_ORDER
        for devtype in devtypes:
            retlist.extend(getattr(self, devtype))
        return retlist


class _IOThreadID(XMLBuilder):
    XML_NAME = "iothread"
    _XML_PROP_ORDER = ["id", "thread_pool_min", "thread_pool_max"]

    id = XMLProperty("./@id", is_int=True)
    thread_pool_min = XMLProperty("./@thread_pool_min", is_int=True)
    thread_pool_max = XMLProperty("./@thread_pool_max", is_int=True)
    max = XMLProperty("./poll/@max", is_int=True)
    grow = XMLProperty("./poll/@grow", is_int=True)
    shrink = XMLProperty("./poll/@shrink", is_int=True)


class _DefaultIOThread(XMLBuilder):
    XML_NAME = "defaultiothread"

    thread_pool_min = XMLProperty("./@thread_pool_min", is_int=True)
    thread_pool_max = XMLProperty("./@thread_pool_max", is_int=True)


class Guest(XMLBuilder):
    @staticmethod
    def validate_name(conn, name, check_collision=True, validate=True):
        if validate:
            XMLBuilder.validate_generic_name(_("Guest"), name)
        if not check_collision:
            return

        try:
            conn.lookupByName(name)
        except Exception:
            return
        raise ValueError(_("Guest name '%s' is already in use.") % name)

    @staticmethod
    def generate_uuid(conn):
        def _randomUUID():
            if conn.fake_conn_predictable():
                # Testing hack
                return "00000000-1111-2222-3333-444444444444"

            u = [random.randint(0, 255) for ignore in range(0, 16)]
            u[6] = (u[6] & 0x0F) | (4 << 4)
            u[8] = (u[8] & 0x3F) | (2 << 6)

            return "-".join(["%02x" * 4, "%02x" * 2, "%02x" * 2, "%02x" * 2, "%02x" * 6]) % tuple(u)

        for ignore in range(256):
            uuid = _randomUUID()
            if not generatename.check_libvirt_collision(conn.lookupByUUID, uuid):
                return uuid

        log.error("Failed to generate non-conflicting UUID")  # pragma: no cover

    @staticmethod
    def generate_name(guest):
        def _pretty_arch(_a):
            if _a == "armv7l":
                return "arm"
            return _a

        force_num = False
        basename = guest.osinfo.name
        if basename.endswith("-unknown"):
            basename = basename.rsplit("-", 1)[0]

        if guest.osinfo.name == "generic":
            force_num = True
            if guest.os.is_container():
                basename = "container"
            else:
                basename = "vm"

        if guest.os.arch != guest.conn.caps.host.cpu.arch:
            basename += "-%s" % _pretty_arch(guest.os.arch)
            force_num = False

        def cb(n):
            return generatename.check_libvirt_collision(guest.conn.lookupByName, n)

        return generatename.generate_name(
            basename,
            cb,
            start_num=force_num and 1 or 2,
            force_num=force_num,
            sep=not force_num and "-" or "",
        )

    @staticmethod
    def get_recommended_machine(capsinfo):
        """
        Return the recommended machine type for the passed capsinfo.
        We only return this for arch cases where there's a very clear
        preference that's different from the default machine type
        """

        def _qemu_machine():
            if capsinfo.arch in ["ppc64", "ppc64le"] and "pseries" in capsinfo.machines:
                return "pseries"

            if capsinfo.arch in ["armv7l", "aarch64"]:
                if "virt" in capsinfo.machines:
                    return "virt"
                if "vexpress-a15" in capsinfo.machines:  # pragma: no cover
                    return "vexpress-a15"

            if capsinfo.arch in ["s390x"]:
                if "s390-ccw-virtio" in capsinfo.machines:
                    return "s390-ccw-virtio"

            if capsinfo.arch in ["riscv64", "riscv32"]:
                if "virt" in capsinfo.machines:
                    return "virt"

        if capsinfo.conn.is_qemu() or capsinfo.conn.is_test():
            return _qemu_machine()
        return None

    #################
    # init handling #
    #################

    XML_NAME = "domain"
    _XML_PROP_ORDER = [
        "type",
        "name",
        "uuid",
        "genid",
        "genid_enable",
        "title",
        "description",
        "_metadata",
        "iothreads",
        "iothreadids",
        "defaultiothread",
        "maxMemory",
        "maxMemorySlots",
        "memory",
        "_currentMemory",
        "blkiotune",
        "memtune",
        "memoryBacking",
        "_vcpus",
        "vcpu_current",
        "vcpu_placement",
        "vcpu_cpuset",
        "vcpulist",
        "numatune",
        "resource",
        "sysinfo",
        "bootloader",
        "bootloader_args",
        "os",
        "idmap",
        "features",
        "cpu",
        "clock",
        "on_poweroff",
        "on_reboot",
        "on_crash",
        "pm",
        "emulator",
        "devices",
        "launchSecurity",
        "seclabels",
        "keywrap",
    ]

    def __init__(self, *args, **kwargs):
        XMLBuilder.__init__(self, *args, **kwargs)

        # Allow virt-manager to override the default graphics type
        self.default_graphics_type = BuildConfig.default_graphics

        self.skip_default_console = False
        self.skip_default_channel = False
        self.skip_default_sound = False
        self.skip_default_usbredir = False
        self.skip_default_graphics = False
        self.skip_default_rng = False
        self.skip_default_tpm = False
        self.skip_default_input = False
        self.have_default_tpm = False
        self.x86_cpu_default = self.cpu.SPECIAL_MODE_APP_DEFAULT

        # qemu 6.1, fairly new when we added this option, has an unfortunate
        # bug with >= 15 root ports, so we choose 14 instead of our original 16
        # https://gitlab.com/qemu-project/qemu/-/issues/641
        self.num_pcie_root_ports = 14

        self.skip_default_osinfo = False
        self.uefi_requested = None
        self.__osinfo = None
        self._capsinfo = None
        self._domcaps = None
        self._extra_drivers = None

    ######################
    # Property accessors #
    ######################

    name = XMLProperty("./name")

    iothreads = XMLProperty("./iothreads", is_int=True)
    iothreadids = XMLChildProperty(_IOThreadID, relative_xpath="./iothreadids")
    defaultiothread = XMLChildProperty(_DefaultIOThread)

    def _set_currentMemory(self, val):
        if val is not None:
            val = int(val)
            if self.memory is None or self.memory < val:
                self.memory = val
        self._currentMemory = val

    def _get_currentMemory(self):
        return self._currentMemory

    currentMemory = property(_get_currentMemory, _set_currentMemory)

    _currentMemory = XMLProperty("./currentMemory", is_int=True)
    memory = XMLProperty("./memory", is_int=True)
    maxMemory = XMLProperty("./maxMemory", is_int=True)
    maxMemorySlots = XMLProperty("./maxMemory/@slots", is_int=True)

    def _set_vcpus(self, val):
        if val is not None:
            val = int(val)
            # Don't force set curvcpus unless already specified
            if self.vcpu_current is not None and self.vcpu_current > val:
                self.vcpu_current = val
        self._vcpus = val

    def _get_vcpus(self):
        return self._vcpus

    _vcpus = XMLProperty("./vcpu", is_int=True)
    vcpus = property(_get_vcpus, _set_vcpus)

    vcpu_current = XMLProperty("./vcpu/@current", is_int=True)
    vcpu_placement = XMLProperty("./vcpu/@placement")
    vcpu_cpuset = XMLProperty("./vcpu/@cpuset")

    uuid = XMLProperty("./uuid")
    genid = XMLProperty("./genid")
    genid_enable = XMLProperty("./genid", is_bool=True)
    id = XMLProperty("./@id", is_int=True)
    type = XMLProperty("./@type")
    bootloader = XMLProperty("./bootloader")
    bootloader_args = XMLProperty("./bootloader_args")
    description = XMLProperty("./description")
    title = XMLProperty("./title")
    emulator = XMLProperty("./devices/emulator")

    on_poweroff = XMLProperty("./on_poweroff")
    on_reboot = XMLProperty("./on_reboot")
    on_crash = XMLProperty("./on_crash")
    on_lockfailure = XMLProperty("./on_lockfailure")

    vcpulist = XMLChildProperty(DomainVCPUs, is_single=True)
    seclabels = XMLChildProperty(DomainSeclabel)
    keywrap = XMLChildProperty(DomainKeyWrap, is_single=True)
    os = XMLChildProperty(DomainOs, is_single=True)
    features = XMLChildProperty(DomainFeatures, is_single=True)
    clock = XMLChildProperty(DomainClock, is_single=True)
    cpu = XMLChildProperty(DomainCpu, is_single=True)
    cputune = XMLChildProperty(DomainCputune, is_single=True)
    numatune = XMLChildProperty(DomainNumatune, is_single=True)
    pm = XMLChildProperty(DomainPm, is_single=True)
    blkiotune = XMLChildProperty(DomainBlkiotune, is_single=True)
    memtune = XMLChildProperty(DomainMemtune, is_single=True)
    memoryBacking = XMLChildProperty(DomainMemoryBacking, is_single=True)
    idmap = XMLChildProperty(DomainIdmap, is_single=True)
    resource = XMLChildProperty(DomainResource, is_single=True)
    sysinfo = XMLChildProperty(DomainSysinfo)
    launchSecurity = XMLChildProperty(DomainLaunchSecurity, is_single=True)
    _metadata = XMLChildProperty(DomainMetadata, is_single=True)

    xmlns_qemu = XMLChildProperty(DomainXMLNSQemu, is_single=True)

    ##############################
    # osinfo related definitions #
    ##############################

    def _get_osinfo(self):
        if self.__osinfo:
            return self.__osinfo

        os_id = self._metadata.libosinfo.os_id
        if os_id:
            self.__osinfo = OSDB.lookup_os_by_full_id(os_id)
            if not self.__osinfo:
                log.debug(
                    "XML had libosinfo os id=%s but we didn't "
                    "find any libosinfo object matching that",
                    os_id,
                )

        if not self.__osinfo:
            # If you hit this error, it means some part of the cli
            # tried to access osinfo before we can depend on it being
            # available. Try moving whatever bits need osinfo to be
            # triggered via set_defaults
            if self.skip_default_osinfo:
                raise xmlutil.DevError("osinfo is accessed before it has been set.")
            self.set_default_os_name()
        return self.__osinfo

    osinfo = property(_get_osinfo)

    def _set_os_obj(self, obj):
        self.__osinfo = obj
        self._metadata.libosinfo.os_id = obj.full_id

    def set_os_name(self, name):
        obj = OSDB.lookup_os(name, raise_error=True)
        log.debug("Setting Guest osinfo name %s", obj)
        self._set_os_obj(obj)

    def set_default_os_name(self):
        self.set_os_name("generic")

    def _supports_virtio(self, os_support):
        if not self.conn.is_qemu():
            return False

        # These _only_ support virtio so don't check the OS
        if (
            self.os.is_riscv_virt()
            or self.os.is_s390x()
            or self.os.is_pseries()
            or self.os.is_loongarch64()
        ):
            return True

        return os_support

    def supports_virtionet(self):
        return self._supports_virtio(self.osinfo.supports_virtionet(self._extra_drivers))

    def supports_virtiodisk(self):
        return self._supports_virtio(self.osinfo.supports_virtiodisk(self._extra_drivers))

    def supports_virtioscsi(self):
        return self._supports_virtio(self.osinfo.supports_virtioscsi(self._extra_drivers))

    def _supports_virtioserial(self):
        return self._supports_virtio(self.osinfo.supports_virtioserial(self._extra_drivers))

    #####################
    # Bootorder helpers #
    #####################

    def _get_old_boot_order(self):
        return self.os.bootorder

    def _convert_old_boot_order(self):
        """Converts the old boot order (e.g. <boot dev='hd'/>) into the
        per-device boot order format.

        """
        boot_order = self._get_old_boot_order()
        ret = []
        disk = None
        cdrom = None
        floppy = None
        net = None

        for d in self.devices.disk:
            if not cdrom and d.device == "cdrom":
                cdrom = d
            if not floppy and d.device == "floppy":
                floppy = d
            if not disk and d.device not in ["cdrom", "floppy"]:
                disk = d
            if cdrom and disk and floppy:
                break

        for n in self.devices.interface:
            net = n
            break

        for b in boot_order:
            if b == "network" and net:
                ret.append(net.get_xml_id())
            elif b == "hd" and disk:
                ret.append(disk.get_xml_id())
            elif b == "cdrom" and cdrom:
                ret.append(cdrom.get_xml_id())
            elif b == "fd" and floppy:
                ret.append(floppy.get_xml_id())
        return ret

    def _get_device_boot_order(self):
        order = []
        for dev in self.get_bootable_devices():
            if not dev.boot.order:
                continue
            order.append((dev.get_xml_id(), dev.boot.order))

        if not order:
            # No devices individually marked bootable, convert traditional
            # boot XML to fine grained
            return self._convert_old_boot_order()

        order.sort(key=lambda p: p[1])
        return [p[0] for p in order]

    def get_boot_order(self, legacy=False):
        if legacy:
            return self._get_old_boot_order()
        return self._get_device_boot_order()

    def _set_device_boot_order(self, boot_order):
        """Sets the new device boot order for the domain"""
        # Unset the traditional boot order
        self.os.bootorder = []

        # Unset device boot order
        for dev in self.devices.get_all():
            dev.boot.order = None

        dev_map = dict((dev.get_xml_id(), dev) for dev in self.get_bootable_devices())
        for boot_idx, dev_xml_id in enumerate(boot_order, 1):
            dev_map[dev_xml_id].boot.order = boot_idx

    def set_boot_order(self, boot_order, legacy=False):
        """Modifies the boot order"""
        if legacy:
            self.os.bootorder = boot_order
        else:
            self._set_device_boot_order(boot_order)

    def reorder_boot_order(self, dev, boot_index):
        """Sets boot order of `dev` to `boot_index`

        Sets the boot order for device `dev` to value `boot_index` and
        adjusts all other boot indices accordingly. Additionally the
        boot order defined in the 'os' node of a domain definition is
        disabled since they are mutually exclusive in libvirt.

        """
        # unset legacy boot order
        self.os.bootorder = []

        # Sort the bootable devices by boot order
        devs_sorted = sorted(
            [device for device in self.get_bootable_devices() if device.boot.order is not None],
            key=lambda device: device.boot.order,
        )

        # set new boot order
        dev.boot.order = boot_index

        next_boot_index = None
        for device in devs_sorted:
            if device is dev:
                continue

            if device.boot.order in [next_boot_index, boot_index]:
                next_boot_index = device.boot.order + 1
                device.boot.order = next_boot_index
                continue

            if next_boot_index is not None:
                # we found a hole so we can stop here
                break

    ###############################
    # Public XML APIs and helpers #
    ###############################

    def add_device(self, dev):
        self.devices.add_child(dev)

    def remove_device(self, dev):
        self.devices.remove_child(dev)
        self._remove_duplicate_console(dev)
        self._remove_spice_devices(dev)

    devices = XMLChildProperty(_DomainDevices, is_single=True)

    def find_device(self, origdev):
        """
        Try to find a child device that matches the content of
        the passed @origdev.
        """
        devlist = getattr(self.devices, origdev.DEVICE_TYPE)
        for idx, dev in enumerate(devlist):
            if origdev.compare_device(dev, idx):
                return dev
        return None

    def get_bootable_devices(self, exclude_redirdev=False):
        """
        Returns bootable devices of the guest definition. If
        @exclude_redirdev is `True` redirected devices will be
        skipped in the output.

        """
        devices = self.devices
        devs = devices.disk + devices.interface + devices.hostdev
        if not exclude_redirdev:
            devs = devs + devices.redirdev
        return devs

    def prefers_uefi(self):
        """
        Return True if this config prefers UEFI. For example,
        arm+machvirt prefers UEFI since it's required for traditional
        install methods
        """
        if self.os.is_x86() and (self.conn.is_qemu() or self.conn.is_test()):
            # If OS has dropped support for 'bios', we have no
            # choice but to use EFI.
            # For other OS still prefer BIOS since it is faster
            # and doesn't break QEMU internal snapshots
            prefer_efi = self.osinfo.requires_firmware_efi(self.os.arch)
        else:
            prefer_efi = (
                self.os.is_arm_machvirt()
                or self.os.is_riscv_virt()
                or self.os.is_loongarch64()
                or self.conn.is_bhyve()
            )

        log.debug("Prefer EFI => %s", prefer_efi)
        return prefer_efi

    def is_uefi(self):
        if self.os.loader and self.os.loader_type == "pflash":
            return True
        return self.os.firmware == "efi"

    def set_uefi_path(self, path):
        """
        Set old style UEFI XML via loader path.
        Set up smm if needed for secureboot
        """
        self.os.loader_ro = True
        self.os.loader_type = "pflash"
        self.os.loader = path

        # If the firmware name contains "secboot" it is probably build
        # with SMM feature required so we need to enable that feature,
        # otherwise the firmware may fail to load.  True secure boot is
        # currently supported only on x86 architecture and with q35 with
        # SMM feature enabled so change the machine to q35 as well.
        # To actually enforce the secure boot for the guest if Secure Boot
        # Mode is configured we need to enable loader secure feature.
        if self.os.is_x86() and "secboot" in self.os.loader:
            self.features.smm = True
            self.os.loader_secure = True
            if not self.os.is_q35():
                log.warning(
                    "Changing machine type from '%s' to 'q35' "
                    "which is required for UEFI secure boot.",
                    self.os.machine,
                )
                self.os.machine = "q35"

    def enable_uefi(self):
        """
        Enable UEFI using our default logic
        """
        domcaps = self.lookup_domcaps()
        if domcaps.supports_firmware_efi():
            self.os.firmware = "efi"
            return

        path = self._lookup_default_uefi_path()
        log.debug("Setting default UEFI path=%s", path)
        self.set_uefi_path(path)

    def disable_uefi(self):
        self.os.firmware = None
        self.os.loader = None
        self.os.loader_ro = None
        self.os.loader_type = None
        self.os.loader_secure = None
        self.os.nvram = None
        self.os.nvram_template = None
        for feature in self.os.firmware_features:
            self.os.remove_child(feature)

        # Force remove any properties we don't know about
        self._xmlstate.xmlapi.node_force_remove("./os/firmware")
        self._xmlstate.xmlapi.node_force_remove("./os/nvram")
        self._xmlstate.xmlapi.node_force_remove("./os/loader")

    def has_spice(self):
        for gfx in self.devices.graphics:
            if gfx.type == gfx.TYPE_SPICE:
                return True

    def has_gl(self):
        for gfx in self.devices.graphics:
            if gfx.gl:
                return True

    def has_listen_none(self):
        for gfx in self.devices.graphics:
            listen = gfx.get_first_listen_type()
            if listen and listen == "none":
                return True
        return False

    def is_full_os_container(self):
        if not self.os.is_container():
            return False
        for fs in self.devices.filesystem:
            if fs.target == "/":
                return True
        return False

    def can_default_virtioscsi(self):
        """
        Return True if the guest supports virtio-scsi, and there's
        no other scsi controllers attached to the guest
        """
        has_any_scsi = any([d.type == "scsi" for d in self.devices.controller])
        return not has_any_scsi and self.supports_virtioscsi()

    def hyperv_supported(self):
        if not self.osinfo.is_windows():
            return False
        return True

    def lookup_domcaps(self):
        def _compare_machine(domcaps):
            if self.os.machine == domcaps.machine:
                return True
            try:
                capsinfo = self.lookup_capsinfo()
            except Exception:
                log.exception("Error fetching machine list for alias resolution, assuming mismatch")
                return False
            if capsinfo.is_machine_alias(self.os.machine, domcaps.machine):
                return True
            return False

        # We need to regenerate domcaps cache if any of these values change
        def _compare(domcaps):
            if self.type == "test":
                # Test driver doesn't support domcaps. We kinda fake it in
                # some cases, but it screws up the checking here for parsed XML
                return True
            if self.os.machine and not _compare_machine(domcaps):
                return False
            if self.type and self.type != domcaps.domain:
                return False
            if self.os.arch and self.os.arch != domcaps.arch:
                return False  # pragma: no cover
            if self.emulator and self.emulator != domcaps.path:
                return False
            return True

        if not self._domcaps or not _compare(self._domcaps):
            self._domcaps = DomainCapabilities.build_from_guest(self)
        return self._domcaps

    def lookup_capsinfo(self):
        # We need to regenerate capsinfo cache if any of these values change
        def _compare(capsinfo):  # pragma: no cover
            if self.type and self.type != capsinfo.hypervisor_type:
                return False
            if self.os.os_type and self.os.os_type != capsinfo.os_type:
                return False
            if self.os.arch and self.os.arch != capsinfo.arch:
                return False
            if self.os.machine and self.os.machine not in capsinfo.machines:
                return False
            return True

        if not self._capsinfo or not _compare(self._capsinfo):
            self._capsinfo = self.conn.caps.guest_lookup(
                os_type=self.os.os_type, arch=self.os.arch, typ=self.type, machine=self.os.machine
            )
        return self._capsinfo

    def set_capabilities_defaults(self, capsinfo=None):
        if capsinfo:  # pragma: no cover
            self._capsinfo = capsinfo
        else:
            capsinfo = self.lookup_capsinfo()
        wants_default_type = not self.type and not self.os.os_type

        self.type = capsinfo.hypervisor_type
        self.os.os_type = capsinfo.os_type
        self.os.arch = capsinfo.arch
        if not self.os.loader:
            self.os.loader = capsinfo.loader
        if not self.emulator and not self.os.is_xenpv() and self.type != "vz":
            self.emulator = capsinfo.emulator
        if not self.os.machine:
            self.os.machine = Guest.get_recommended_machine(capsinfo)

        if wants_default_type and self.conn.is_qemu() and self.os.is_x86() and self.type != "kvm":
            log.warning("KVM acceleration not available, using '%s'", self.type)  # pragma: no cover

    def refresh_machine_type(self):
        """
        Reset the guests's machine type to the latest 'canonical' machine
        name that qemu reports. So if my VM is using ancient pc-0.11, we
        try to turn that into just `pc`

        The algorithm here is to fetch all machine types that are aliases
        for a stable name (like pc -> pc-i440fx-6.2), and see if our current
        machine type uses alias as a prefix. This is the format that qemu
        uses for its stable machine type names.
        """
        # We need to unset the machine type first, so we can perform
        # a successful capsinfo lookup, otherwise we will error when qemu
        # has deprecated and removed the old machine type
        original_machine_type = self.os.machine or ""
        self.os.machine = None

        capsinfo = self.lookup_capsinfo()
        mobjs = (capsinfo.domain and capsinfo.domain.machines) or capsinfo.guest.machines
        canonical_names = [m.name for m in mobjs if m.canonical]

        for machine_alias in canonical_names:
            if machine_alias == "pc":
                prefix = "pc-i440fx-"
            elif machine_alias == "q35":
                prefix = "pc-q35-"
            else:
                # Example: pseries-X, virt-X, s390-ccw-virtio-X
                prefix = machine_alias + "-"

            if original_machine_type.startswith(prefix):
                self.os.machine = machine_alias
                return
        raise RuntimeError("Don't know how to refresh machine type '%s'" % original_machine_type)

    def set_smbios_serial_cloudinit(self):
        if not self.conn.is_qemu() and not self.conn.is_test():
            return  # pragma: no cover
        if not self.os.is_x86() and not self.os.is_arm_machvirt():
            return  # pragma: no cover
        if self.os.smbios_mode not in [None, "sysinfo"]:
            return

        sysinfos = [s for s in self.sysinfo if s.type == "smbios"]
        if not sysinfos:
            sysinfos = [self.sysinfo.add_new()]
        sysinfo = sysinfos[0]

        if sysinfo.system_serial:
            return
        self.os.smbios_mode = "sysinfo"
        sysinfo.type = "smbios"
        sysinfo.system_serial = "ds=nocloud"

    def sync_vcpus_topology(self, defCPUs):
        """
        <cpu> topology count and <vcpus> always need to match. Handle
        the syncing here since we are less constrained then doing it
        in CPU set_defaults
        """
        if not self.vcpus:
            if self.cpu.has_topology():
                self.vcpus = self.cpu.vcpus_from_topology()
            else:
                self.vcpus = defCPUs
        self.cpu.set_topology_defaults(self.vcpus)

    def change_graphics(self, val, inst):
        inst.type = val
        if val == "spice":
            self._add_spice_devices()
        else:
            self._remove_spice_devices(inst)

    def convert_to_q35(self, num_pcie_root_ports=None):
        self.os.machine = "q35"

        if num_pcie_root_ports is not None:
            self.num_pcie_root_ports = int(num_pcie_root_ports)

        for dev in self.devices.get_all():
            if dev.DEVICE_TYPE == "controller" and dev.type in ["pci", "ide"]:
                self.remove_device(dev)
                continue

            if dev.DEVICE_TYPE == "sound" and dev.model == "ich6":
                dev.model = "ich9"

            if dev.DEVICE_TYPE == "interface" and dev.model == "e1000":
                dev.model = "e1000e"

            if dev.DEVICE_TYPE == "disk" and dev.bus == "ide":
                dev.bus = "sata"
                used_targets = [d.target for d in self.devices.disk if d.target]
                dev.generate_target(used_targets)
                dev.address.clear()

            if dev.address.type == "pci":
                dev.address.clear()

        self.add_q35_pcie_controllers()

    def _convert_spice_gl_to_egl_headless(self):
        if not self.has_spice():
            return

        spicedev = [g for g in self.devices.graphics if g.type == "spice"][0]
        if not spicedev.gl:
            return

        dev = DeviceGraphics(self.conn)
        dev.type = "egl-headless"
        dev.set_defaults(self.conn)
        if spicedev.rendernode:
            dev.rendernode = spicedev.rendernode
        self.add_device(dev)

    def _convert_to_vnc_graphics(self, agent):
        """
        If there's already VNC graphics configured, we leave it intact,
        but rip out all evidence of other graphics devices.

        If there's other non-VNC, non-egl-headless configured, we try to
        inplace convert the first device we encounter.

        If there's no graphics configured, set up a default VNC config.`
        """
        vnc_devs = [g for g in self.devices.graphics if g.type == "vnc"]
        # We ignore egl-headless, it's not a true graphical frontend
        other_devs = [
            g for g in self.devices.graphics if g.type != "vnc" and g.type != "egl-headless"
        ]

        # Guest already had a vnc device.
        # Remove all other devs and we are done
        if vnc_devs:
            for g in other_devs:
                self.remove_device(g)
            return

        # We didn't find any non-vnc device to convert.
        # Add a vnc device with default config
        if not other_devs:
            dev = DeviceGraphics(self.conn)
            dev.type = dev.TYPE_VNC
            dev.set_defaults(self.conn)
            self.add_device(dev)
            return

        # Convert the pre-existing graphics device to vnc
        # Remove the rest
        dev = other_devs.pop(0)
        srcdev = DeviceGraphics(self.conn, dev.get_xml())
        for g in other_devs:
            self.remove_device(g)

        dev.clear()
        dev.type = dev.TYPE_VNC
        dev.keymap = srcdev.keymap
        dev.port = srcdev.port
        dev.autoport = srcdev.autoport
        dev.passwd = srcdev.passwd
        dev.passwdValidTo = srcdev.passwdValidTo
        dev.listen = srcdev.listen
        for listen in srcdev.listens:
            srcdev.remove_child(listen)
            dev.add_child(listen)
        dev.set_defaults(self)

        if agent:
            agent.source.clipboard_copypaste = srcdev.clipboard_copypaste
            agent.source.mouse_mode = srcdev.mouse_mode

    def _convert_to_vnc_video(self):
        """
        If there's no video device, add a default one.
        If there's any qxl device, reset its config to app defaults.
        """
        if not self.devices.video:
            videodev = DeviceVideo(self.conn)
            videodev.set_defaults(self)
            self.add_device(videodev)
            return

        qxl_devs = [v for v in self.devices.video if v.model == "qxl"]
        if qxl_devs and not any(dev.primary for dev in self.devices.video):
            # Make sure `primary` flag is set, we need it up ahead
            self.devices.video[0].primary = True

        for dev in qxl_devs:
            is_primary = dev.primary
            dev.clear()
            dev.set_defaults(self)
            if not is_primary and dev.model != "virtio":
                # Device can't be non-primary, so just remove it
                log.debug(
                    "Can't use model=%s for non-primary video device, removing it instead.",
                    dev.model,
                )
                self.remove_device(dev)

    def _add_qemu_vdagent(self):
        if any(c for c in self.devices.channel if c.type == c.TYPE_QEMUVDAGENT):
            return

        dev = DeviceChannel(self.conn)
        dev.type = dev.TYPE_QEMUVDAGENT
        dev.set_defaults(self)
        self.add_device(dev)
        return dev

    def convert_to_vnc(self, qemu_vdagent=False):
        """
        Convert existing XML to have one VNC graphics connection.
        """
        self._convert_spice_gl_to_egl_headless()

        # Rip out spice graphics devices unconditionally.
        # Could be necessary if XML is in broken state.
        self._force_remove_spice_devices()

        agent = None
        if qemu_vdagent:
            agent = self._add_qemu_vdagent()

        self._convert_to_vnc_graphics(agent)
        self._convert_to_vnc_video()

    def set_defaults(self, _guest):
        self.set_capabilities_defaults()

        if not self.uuid:
            self.uuid = Guest.generate_uuid(self.conn)

        self.sync_vcpus_topology(1)

        self._set_default_machine()
        self._set_default_uefi()

        self._add_default_graphics()
        self._add_default_video_device()
        self._add_default_input_device()
        self._add_default_console_device()
        self._add_default_usb_controller()
        self._add_default_channels()
        self._add_default_rng()
        self._add_default_memballoon()
        self._add_default_tpm()

        self.clock.set_defaults(self)
        self.cpu.set_defaults(self)
        self.features.set_defaults(self)
        for seclabel in self.seclabels:
            seclabel.set_defaults(self)
        self.pm.set_defaults(self)
        self.os.set_defaults(self)
        self.launchSecurity.set_defaults(self)

        for dev in self.devices.get_all():
            dev.set_defaults(self)

        self.add_virtioscsi_controller()
        self.add_q35_pcie_controllers()
        self._add_spice_devices()

    def add_extra_drivers(self, extra_drivers):
        self._extra_drivers = extra_drivers

    ########################
    # Private xml routines #
    ########################

    def _set_default_machine(self):
        if self.os.machine:
            return

        capsinfo = self.lookup_capsinfo()

        if (
            self.os.is_x86()
            and self.conn.is_qemu()
            and "q35" in capsinfo.machines
            and self.conn.support.qemu_q35_default()
            and self.osinfo.supports_chipset_q35()
        ):
            self.os.machine = "q35"
            return

        default = capsinfo.machines and capsinfo.machines[0] or None
        self.os.machine = default

    def _lookup_default_uefi_path(self):
        """
        If a default UEFI firmware path is found, return it,
        otherwise raise an error
        """
        domcaps = self.lookup_domcaps()

        if not domcaps.supports_uefi_loader():
            raise RuntimeError(_("Libvirt version does not support UEFI."))

        if not domcaps.arch_can_uefi():
            raise RuntimeError(  # pragma: no cover
                _("Don't know how to setup UEFI for arch '%s'") % self.os.arch
            )

        path = domcaps.find_uefi_path_for_arch()
        if not path:  # pragma: no cover
            raise RuntimeError(_("Did not find any UEFI binary path for arch '%s'") % self.os.arch)

        return path

    def _set_default_uefi(self):
        if self.uefi_requested is False:
            return

        use_default_uefi = (
            self.prefers_uefi()
            and not self.os.kernel
            and not self.os.loader
            and self.os.loader_ro is None
            and self.os.nvram is None
            and self.os.firmware is None
        )

        if not use_default_uefi and not self.uefi_requested:
            return

        try:
            self.enable_uefi()
        except RuntimeError as e:
            if self.uefi_requested:
                raise
            log.debug("Error setting UEFI default", exc_info=True)
            log.warning("Couldn't configure UEFI: %s", e)
            log.warning("Your VM may not boot successfully.")

    def _usb_disabled(self):
        controllers = [c for c in self.devices.controller if c.type == "usb"]
        if not controllers:
            return False
        return all([c.model == "none" for c in controllers])

    def _add_default_input_device(self):
        if self.skip_default_input:
            return
        if self.os.is_container():
            return
        if self.os.is_xenpv():
            return
        if self.devices.input:
            return
        if not self.devices.graphics:
            return

        tablet = True
        keyboard = True
        if self.os.is_x86():
            # We don't historically add USB keyboard for x86,
            # default libvirt/qemu PS2 seems to be fine
            keyboard = False

        bus = None
        if self.os.is_s390x():
            # s390x guests need VirtIO input devices
            if self.osinfo.supports_virtioinput(self._extra_drivers):
                bus = "virtio"
        elif not self._usb_disabled():
            bus = "usb"

        if not bus:
            return

        def _add_input(itype):
            dev = DeviceInput(self.conn)
            dev.type = itype
            dev.bus = bus
            dev.set_defaults(self)
            self.add_device(dev)

        if tablet:
            _add_input("tablet")
        if keyboard:
            _add_input("keyboard")

    def _add_default_console_device(self):
        if self.skip_default_console:
            return
        if self.devices.console or self.devices.serial:
            return

        dev = DeviceConsole(self.conn)
        if self.conn.is_bhyve():
            nmdm_dev_prefix = "/dev/nmdm{}".format(self.generate_uuid(self.conn))
            dev.type = dev.TYPE_NMDM
            dev.source.master = nmdm_dev_prefix + "A"
            dev.source.slave = nmdm_dev_prefix + "B"
        else:
            dev.type = dev.TYPE_PTY

        if self.os.is_s390x():
            dev.target_type = "sclp"
        self.add_device(dev)

    def _add_default_video_device(self):
        if self.os.is_container():
            return
        if self.devices.video:
            return
        if not self.devices.graphics:
            return
        self.add_device(DeviceVideo(self.conn))

    def _add_default_usb_controller(self):
        if any([d.type == "usb" for d in self.devices.controller]):
            return
        if not self.conn.is_qemu() and not self.conn.is_test():
            return

        qemu_usb3 = self.conn.support.conn_qemu_xhci()
        usb2 = False
        usb3 = False
        if self.os.is_x86():
            usb3 = bool(self.osinfo.supports_usb3() and qemu_usb3)
            usb2 = not usb3
        elif self.os.is_arm_machvirt():
            # For machvirt, we always assume OS supports usb3
            if qemu_usb3 and self.conn.support.conn_machvirt_pci_default():
                usb3 = True
        elif self.os.is_riscv_virt():
            # For RISC-V we can assume the guest OS supports USB3, but we
            # have to make sure libvirt and QEMU are new enough to be using
            # PCI by default
            if qemu_usb3 and self.conn.support.conn_riscv_virt_pci_default():
                usb3 = True
        elif self.os.is_pseries():
            # For pseries, we always assume OS supports usb3
            if qemu_usb3:
                usb3 = True
        elif self.os.is_loongarch64():
            # For loongarch64, we always assume OS supports usb3
            if qemu_usb3:
                usb3 = True

        if usb2:
            for dev in DeviceController.get_usb2_controllers(self.conn):
                self.add_device(dev)
        elif usb3:
            self.add_device(DeviceController.get_usb3_controller(self.conn, self))

    def _add_default_channels(self):
        if self.skip_default_channel:
            return
        if self.devices.channel:
            return

        if (
            self.conn.is_qemu()
            and self._supports_virtioserial()
            and self.conn.support.conn_autosocket()
        ):
            dev = DeviceChannel(self.conn)
            dev.type = "unix"
            dev.target_type = "virtio"
            dev.target_name = dev.CHANNEL_NAME_QEMUGA
            self.add_device(dev)

    def _add_default_graphics(self):
        if self.skip_default_graphics:
            return
        if self.devices.graphics:
            return
        if self.os.is_container() and not self.conn.is_vz():
            return
        if (
            not self.os.is_x86()
            and not self.os.is_pseries()
            and not self.os.is_loongarch64()
            and not self.os.is_arm_machvirt()
            and not self.os.is_riscv_virt()
        ):
            return
        self.add_device(DeviceGraphics(self.conn))

    def _add_default_rng(self):
        if self.skip_default_rng:
            return
        if self.devices.rng:
            return
        if not (
            self.os.is_x86()
            or self.os.is_arm_machvirt()
            or self.os.is_riscv_virt()
            or self.os.is_s390x()
            or self.os.is_pseries()
            or self.os.is_loongarch64()
        ):
            return

        if (
            self.conn.is_qemu()
            and self.osinfo.supports_virtiorng(self._extra_drivers)
            and self.conn.support.conn_rng_urandom()
        ):
            dev = DeviceRng(self.conn)
            dev.type = "random"
            dev.device = "/dev/urandom"
            self.add_device(dev)

    def _add_default_tpm(self):
        if self.skip_default_tpm:
            return
        if self.devices.tpm:
            return

        # If the guest is using UEFI, we take that as a
        # flag that the VM is targeting a modern platform
        # and thus we should also provide an emulated TPM.
        if not self.is_uefi():
            return

        if not self.lookup_domcaps().supports_tpm_emulator():
            log.debug("Domain caps doesn't report TPM support")
            return

        log.debug("Adding default TPM")
        dev = DeviceTpm(self.conn)
        dev.type = DeviceTpm.TYPE_EMULATOR
        self.add_device(dev)
        self.have_default_tpm = True

    def _add_default_memballoon(self):
        if self.devices.memballoon:
            return
        if not self.conn.is_qemu():
            return

        # We know for certain that a memballoon is good to have with these
        # machine types; for other machine types, we leave the decision up
        # to libvirt
        if not (
            self.os.is_x86()
            or self.os.is_arm_machvirt()
            or self.os.is_riscv_virt()
            or self.os.is_s390x()
            or self.os.is_pseries()
            or self.os.is_loongarch64()
        ):
            return

        if self.osinfo.supports_virtioballoon(self._extra_drivers):
            dev = DeviceMemballoon(self.conn)
            dev.model = "virtio"
            self.add_device(dev)

    def add_virtioscsi_controller(self):
        if not self.can_default_virtioscsi():
            return
        if not any([d for d in self.devices.disk if d.bus == "scsi"]):
            return

        ctrl = DeviceController(self.conn)
        ctrl.type = "scsi"
        ctrl.model = "virtio-scsi"
        ctrl.set_defaults(self)
        self.add_device(ctrl)

    def defaults_to_pcie(self):
        if self.os.is_q35():
            return True
        if self.os.is_arm_machvirt():
            return True
        if self.os.is_riscv_virt():
            return True
        if self.os.is_loongarch64():
            return True
        return False

    def add_q35_pcie_controllers(self):
        if any([c for c in self.devices.controller if c.type == "pci"]):
            return
        if not self.defaults_to_pcie():
            return

        added = False
        log.debug("Using num_pcie_root_ports=%s", self.num_pcie_root_ports)
        for dummy in range(max(self.num_pcie_root_ports, 0)):
            if not added:
                # Libvirt forces pcie-root to come first
                ctrl = DeviceController(self.conn)
                ctrl.type = "pci"
                ctrl.model = "pcie-root"
                ctrl.set_defaults(self)
                self.add_device(ctrl)
                added = True
            ctrl = DeviceController(self.conn)
            ctrl.type = "pci"
            ctrl.model = "pcie-root-port"
            ctrl.set_defaults(self)
            self.add_device(ctrl)

    def _add_spice_channels(self):
        if not self.lookup_domcaps().supports_channel_spicevmc():
            return  # pragma: no cover
        if self.skip_default_channel:
            return
        for chn in self.devices.channel:
            if chn.type == chn.TYPE_SPICEVMC:
                return

        # We explicitly don't check for virtioserial support here.
        # We did that for a while, which excluded windows, and
        # we received some complaints.
        # https://bugzilla.redhat.com/show_bug.cgi?id=1660123
        dev = DeviceChannel(self.conn)
        dev.type = DeviceChannel.TYPE_SPICEVMC
        dev.set_defaults(self)
        self.add_device(dev)

    def _add_spice_sound(self):
        if self.skip_default_sound:
            return
        if self.devices.sound:
            return
        if not self.os.is_hvm():
            return  # pragma: no cover

        dev = DeviceSound(self.conn)
        dev.set_defaults(self)
        self.add_device(dev)

    def _add_spice_usbredir(self):
        if not self.lookup_domcaps().supports_redirdev_usb():
            return  # pragma: no cover
        if self.skip_default_usbredir:
            return
        if self.devices.redirdev:
            return
        if self._usb_disabled():
            return
        if not self.os.is_x86():
            return

        # If we use 4 devices here, we fill up all the emulated USB2 slots,
        # and directly assigned devices are forced to fall back to USB1
        # https://bugzilla.redhat.com/show_bug.cgi?id=1135488
        for ignore in range(2):
            dev = DeviceRedirdev(self.conn)
            dev.bus = "usb"
            dev.type = "spicevmc"
            dev.set_defaults(self)
            self.add_device(dev)

    def _add_spice_devices(self):
        if not self.has_spice():
            return

        if self.features.vmport is None and self.os.is_x86() and self.conn.support.conn_vmport():
            self.features.vmport = False

        self._add_spice_channels()
        self._add_spice_sound()
        self._add_spice_usbredir()

    def _remove_duplicate_console(self, dev):
        condup = DeviceConsole.get_console_duplicate(self, dev)
        if condup:
            log.debug("Found duplicate console device:\n%s", condup.get_xml())
            self.devices.remove_child(condup)

    def _remove_spice_audio(self):
        for audio in self.devices.audio:
            if audio.type == "spice":
                self.devices.remove_child(audio)

    def _remove_spice_channels(self):
        for channel in self.devices.channel:
            if channel.type == DeviceChannel.TYPE_SPICEVMC:
                self.devices.remove_child(channel)

    def _remove_spice_usbredir(self):
        for redirdev in self.devices.redirdev:
            if redirdev.type == "spicevmc":
                self.devices.remove_child(redirdev)

    def _remove_spiceport(self):
        for dev in self.devices.serial + self.devices.console:
            if dev.type == dev.TYPE_SPICEPORT:
                self.devices.remove_child(dev)

    def _force_remove_spice_devices(self):
        self._remove_spice_audio()
        self._remove_spice_channels()
        self._remove_spice_usbredir()
        self._remove_spiceport()

    def _remove_spice_devices(self, rmdev):
        if rmdev.DEVICE_TYPE != "graphics" or self.has_spice():
            return
        self._force_remove_spice_devices()



#
# Copyright (C) 2013 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.
#

from .logger import log


def _new_poll_helper(origmap, typename, list_cb, build_cb, support_cb):
    """
    Helper for new style listAll* APIs
    """
    current = {}
    new = {}
    objs = []

    try:
        if support_cb():
            objs = list_cb()
    except Exception as e:  # pragma: no cover
        log.debug("Unable to list all %ss: %s", typename, e)

    for obj in objs:
        name = obj.name()

        if name not in origmap:
            # Object is brand new this period
            current[name] = build_cb(obj, name)
            new[name] = current[name]
        else:
            # Previously known object
            current[name] = origmap[name]
            del origmap[name]

    return (list(origmap.values()), list(new.values()), list(current.values()))


def fetch_nets(backend, origmap, build_cb):
    typename = "network"
    list_cb = backend.listAllNetworks
    support_cb = backend.support.conn_network
    return _new_poll_helper(origmap, typename, list_cb, build_cb, support_cb)


def fetch_pools(backend, origmap, build_cb):
    typename = "pool"
    list_cb = backend.listAllStoragePools
    support_cb = backend.support.conn_storage
    return _new_poll_helper(origmap, typename, list_cb, build_cb, support_cb)


def fetch_volumes(backend, pool, origmap, build_cb):
    typename = "volume"
    list_cb = pool.listAllVolumes
    support_cb = backend.support.conn_storage
    return _new_poll_helper(origmap, typename, list_cb, build_cb, support_cb)


def fetch_nodedevs(backend, origmap, build_cb):
    typename = "nodedev"
    list_cb = backend.listAllDevices
    support_cb = backend.support.conn_nodedev
    return _new_poll_helper(origmap, typename, list_cb, build_cb, support_cb)


def fetch_vms(backend, origmap, build_cb):
    typename = "domain"
    list_cb = backend.listAllDomains
    support_cb = backend.support.conn_domain
    return _new_poll_helper(origmap, typename, list_cb, build_cb, support_cb)



# Copyright (C) 2007, 2013-2014 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os
import sys
import textwrap
import traceback

from gi.repository import Gtk

from virtinst import log

from .baseclass import vmmGObject


def _launch_dialog(dialog, primary_text, secondary_text, title, widget=None, modal=True):
    def fix_text(t):
        if not t:
            return t
        if len(t) > 512:
            t = t[:512] + "..."
        retlines = []
        for line in t.splitlines():
            if not line:
                retlines.append("")
            else:
                retlines.extend(textwrap.wrap(line, 80))
        return "\n".join(retlines)

    primary_text = fix_text(primary_text)
    secondary_text = fix_text(secondary_text)

    dialog.set_property("text", primary_text)
    dialog.format_secondary_text(secondary_text or None)
    dialog.set_title(title)

    if widget:
        dialog.get_content_area().add(widget)

    res = False
    if modal:
        res = dialog.run()
        res = bool(res in [Gtk.ResponseType.YES, Gtk.ResponseType.OK])
        dialog.destroy()
    else:

        def response_destroy(src, ignore):
            src.destroy()

        dialog.connect("response", response_destroy)
        dialog.show()

    return res


class vmmErrorDialog(vmmGObject):
    # singleton instance for non-UI classes
    @classmethod
    def get_instance(cls):
        if not cls._instance:
            cls._instance = vmmErrorDialog(None)
        return cls._instance

    def __init__(self, parent):
        vmmGObject.__init__(self)
        self._parent = parent
        self._simple = None

        self._modal_default = False

    def _cleanup(self):
        pass  # pragma: no cover

    def set_modal_default(self, val):
        self._modal_default = val

    def get_parent(self):
        return self._parent

    def show_err(
        self,
        summary,
        details=None,
        title="",
        modal=None,
        debug=True,
        dialog_type=Gtk.MessageType.ERROR,
        buttons=Gtk.ButtonsType.CLOSE,
        text2=None,
    ):
        if modal is None:
            modal = self._modal_default

        if details is None:
            details = summary
            if sys.exc_info()[0] is not None:
                details += "\n\n" + "".join(traceback.format_exc()).strip()
        else:
            details = str(details)

        if debug:
            debugmsg = "error dialog message:\nsummary=%s" % summary
            if details and details != summary:
                det = details
                if details.startswith(summary):
                    det = details[len(summary) :].strip()
                debugmsg += "\ndetails=%s" % det
            log.debug(debugmsg)

        # Make sure we have consistent details for error dialogs
        if dialog_type == Gtk.MessageType.ERROR and summary not in details:
            details = summary + "\n\n" + details

        dialog = _errorDialog(
            parent=self.get_parent(), flags=0, message_type=dialog_type, buttons=buttons
        )

        return dialog.show_dialog(
            primary_text=summary, secondary_text=text2, details=details, title=title, modal=modal
        )

    ###################################
    # Simple one shot message dialogs #
    ###################################

    def _simple_dialog(self, dialog_type, buttons, text1, text2, title, widget=None, modal=True):

        dialog = Gtk.MessageDialog(
            self.get_parent(),
            flags=Gtk.DialogFlags.DESTROY_WITH_PARENT,
            message_type=dialog_type,
            buttons=buttons,
        )
        if self._simple:
            self._simple.destroy()
        self._simple = dialog
        self._simple.get_accessible().set_name("vmm dialog")

        return _launch_dialog(
            self._simple, text1, text2 or "", title or "", widget=widget, modal=modal
        )

    def val_err(self, text1, text2=None, title=_("Input Error"), modal=True):
        logtext = _("Validation Error: %s") % text1
        if text2:
            logtext += " %s" % text2

        if isinstance(text1, Exception) or isinstance(text2, Exception):
            log.exception(logtext)
        else:
            self._logtrace(logtext)

        dtype = Gtk.MessageType.ERROR
        buttons = Gtk.ButtonsType.OK
        self._simple_dialog(
            dtype, buttons, str(text1), text2 and str(text2) or "", str(title), None, modal
        )
        return False

    def show_info(
        self, text1, text2=None, title="", widget=None, modal=True, buttons=Gtk.ButtonsType.OK
    ):
        dtype = Gtk.MessageType.INFO
        self._simple_dialog(dtype, buttons, text1, text2, title, widget, modal)
        return False

    def yes_no(self, text1, text2=None, title=None):
        dtype = Gtk.MessageType.WARNING
        buttons = Gtk.ButtonsType.YES_NO
        return self._simple_dialog(dtype, buttons, text1, text2, title)

    def ok_cancel(self, text1, text2=None, title=None):
        dtype = Gtk.MessageType.WARNING
        buttons = Gtk.ButtonsType.OK_CANCEL
        return self._simple_dialog(dtype, buttons, text1, text2, title)

    def confirm_unapplied_changes(self):
        """
        Helper function for confirming whether to apply unapplied changes
        """
        return self.chkbox_helper(
            self.config.get_confirm_unapplied,
            self.config.set_confirm_unapplied,
            text1=(_("There are unapplied changes. Would you like to apply them now?")),
            chktext=_("Don't warn me again."),
            default=False,
        )

    ##########################################
    # One shot dialog with a checkbox prompt #
    ##########################################

    def warn_chkbox(self, text1, text2=None, chktext=None, buttons=None):
        dtype = Gtk.MessageType.WARNING
        buttons = buttons or Gtk.ButtonsType.OK_CANCEL
        chkbox = _errorDialog(
            parent=self.get_parent(), flags=0, message_type=dtype, buttons=buttons
        )
        return chkbox.show_dialog(primary_text=text1, secondary_text=text2, chktext=chktext)

    def err_chkbox(self, text1, text2=None, chktext=None, buttons=None):
        dtype = Gtk.MessageType.ERROR
        buttons = buttons or Gtk.ButtonsType.OK
        chkbox = _errorDialog(
            parent=self.get_parent(), flags=0, message_type=dtype, buttons=buttons
        )
        return chkbox.show_dialog(primary_text=text1, secondary_text=text2, chktext=chktext)

    def chkbox_helper(
        self, getcb, setcb, text1, text2=None, default=True, chktext=_("Don't ask me again")
    ):
        """
        Helper to prompt user about proceeding with an operation
        Returns True if the 'yes' or 'ok' button was selected, False otherwise

        @default: What value to return if getcb tells us not to prompt
        """
        do_prompt = getcb()
        if not do_prompt:
            return default

        # pylint: disable=unpacking-non-sequence
        res = self.warn_chkbox(
            text1=text1, text2=text2, chktext=chktext, buttons=Gtk.ButtonsType.YES_NO
        )
        response, skip_prompt = res
        setcb(not skip_prompt)

        return response

    def browse_local(
        self,
        dialog_name,
        start_folder=None,
        _type=None,
        dialog_type=None,
        choose_label=None,
        default_name=None,
        confirm_overwrite=False,
    ):
        """
        Helper function for launching a filechooser

        @dialog_name: String to use in the title bar of the filechooser.
        @start_folder: Folder the filechooser is viewing at startup
        @_type: File extension to filter by (e.g. "iso", "png")
        @dialog_type: Maps to FileChooserDialog 'action'
        """
        if dialog_type is None:
            dialog_type = Gtk.FileChooserAction.OPEN

        fcdialog = Gtk.FileChooserNative.new(
            title=dialog_name,
            parent=self.get_parent(),
            action=dialog_type,
            accept_label=choose_label,
        )

        if default_name:
            fcdialog.set_current_name(default_name)

        fcdialog.set_do_overwrite_confirmation(confirm_overwrite)

        # Set file match pattern (ex. *.png)
        if _type is not None:
            pattern = _type
            name = None
            if isinstance(_type, tuple):
                pattern = _type[0]
                name = _type[1]

            f = Gtk.FileFilter()
            f.add_pattern("*." + pattern)
            if name:
                f.set_name(name)
            fcdialog.set_filter(f)

        if start_folder is not None:
            if os.access(start_folder, os.R_OK):
                fcdialog.set_current_folder(start_folder)

        # Run the dialog and parse the response
        ret = None
        if fcdialog.run() == Gtk.ResponseType.ACCEPT:
            ret = fcdialog.get_filename()
        fcdialog.destroy()

        return ret


class _errorDialog(Gtk.MessageDialog):
    """
    Custom error dialog with optional check boxes or details drop down
    """

    def __init__(self, *args, **kwargs):
        Gtk.MessageDialog.__init__(self, *args, **kwargs)

        self.set_title("")
        for child in self.get_message_area().get_children():
            if hasattr(child, "set_max_width_chars"):
                child.set_max_width_chars(40)

        self.get_accessible().set_name("vmm dialog")

        self.chk_vbox = None
        self.init_chkbox()

        self.buffer = None
        self.buf_expander = None
        self.init_details()

    def init_chkbox(self):
        # Init check items
        self.chk_vbox = Gtk.VBox(False, False)
        self.chk_vbox.set_spacing(0)

        self.chk_vbox.show_all()
        self.vbox.pack_start(self.chk_vbox, False, False, 0)  # pylint: disable=no-member

    def init_details(self):
        # Init details buffer
        self.buffer = Gtk.TextBuffer()
        self.buf_expander = Gtk.Expander.new(_("Details"))
        sw = Gtk.ScrolledWindow()
        sw.set_shadow_type(Gtk.ShadowType.IN)
        sw.set_size_request(400, 240)
        sw.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        details = Gtk.TextView.new_with_buffer(self.buffer)
        details.set_editable(False)
        details.set_overwrite(False)
        details.set_cursor_visible(False)
        details.set_wrap_mode(Gtk.WrapMode.WORD)
        details.set_border_width(6)
        sw.add(details)
        self.buf_expander.add(sw)
        self.vbox.pack_start(self.buf_expander, False, False, 0)  # pylint: disable=no-member
        self.buf_expander.show_all()

    def show_dialog(
        self, primary_text, secondary_text="", title="", details="", chktext="", modal=True
    ):
        chkbox = None
        res = None

        # Hide starting widgets
        self.hide()
        self.buf_expander.hide()
        for c in self.chk_vbox.get_children():
            self.chk_vbox.remove(c)  # pragma: no cover

        if details:
            self.buffer.set_text(details)
            title = title or ""
            self.buf_expander.show()

        if chktext:
            chkbox = Gtk.CheckButton(chktext)
            self.chk_vbox.add(chkbox)
            chkbox.show()

        res = _launch_dialog(self, primary_text, secondary_text or "", title, modal=modal)

        if chktext:
            res = [res, chkbox.get_active()]

        return res



# Copyright (C) 2006, 2013, 2014 Red Hat, Inc.
# Copyright (C) 2006 Daniel P. Berrange <berrange@redhat.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os
import time
import threading

import libvirt

from virtinst import DeviceConsole
from virtinst import DeviceController
from virtinst import DeviceDisk
from virtinst import DomainSnapshot
from virtinst import Guest
from virtinst import log

from .libvirtobject import vmmLibvirtObject
from ..baseclass import vmmGObject
from ..lib.libvirtenummap import LibvirtEnumMap
from ..lib import testmock


class _SENTINEL:
    pass


def start_job_progress_thread(vm, meter, progtext):
    current_thread = threading.current_thread()

    def jobinfo_cb():
        while True:
            time.sleep(0.5)

            if not current_thread.is_alive():
                return

            try:
                jobinfo = vm.job_info()
                data_total = int(jobinfo[3])
                data_remaining = int(jobinfo[5])

                # data_total is 0 if the job hasn't started yet
                if not data_total:
                    continue  # pragma: no cover

                if not meter.is_started():
                    meter.start(progtext, data_total)

                progress = data_total - data_remaining
                meter.update(progress)
            except Exception:  # pragma: no cover
                log.exception("Error calling jobinfo")
                return

    if vm.supports_domain_job_info():
        t = threading.Thread(target=jobinfo_cb, name="job progress reporting", args=())
        t.daemon = True
        t.start()


class _IPFetcher:
    """
    Helper class to contain all IP fetching and processing logic
    """

    def __init__(self):
        self._cache = None

    def refresh(self, vm, iface):
        self._cache = {"qemuga": {}, "arp": {}}

        if iface.type == "network":
            net = vm.conn.get_net_by_name(iface.source)
            if net:
                net.get_dhcp_leases(refresh=True)

        if not vm.is_active():
            return

        if vm.agent_ready():
            self._cache["qemuga"] = vm.get_interface_addresses(
                iface, libvirt.VIR_DOMAIN_INTERFACE_ADDRESSES_SRC_AGENT
            )

        arp_flag = getattr(libvirt, "VIR_DOMAIN_INTERFACE_ADDRESSES_SRC_ARP", 3)
        self._cache["arp"] = vm.get_interface_addresses(iface, arp_flag)

    def get(self, vm, iface):
        if self._cache is None:
            self.refresh(vm, iface)

        qemuga = self._cache["qemuga"]
        arp = self._cache["arp"]
        leases = []
        if iface.type == "network":
            net = vm.conn.get_net_by_name(iface.source)
            if net:
                leases = net.get_dhcp_leases()

        def extract_dom(addrs):
            ipv4 = None
            ipv6 = None
            if addrs["hwaddr"] == iface.macaddr:
                for addr in addrs["addrs"] or []:
                    if addr["type"] == 0:
                        ipv4 = addr["addr"]
                    elif addr["type"] == 1 and not str(addr["addr"]).startswith("fe80"):
                        ipv6 = addr["addr"] + "/" + str(addr["prefix"])
            return ipv4, ipv6

        def extract_lease(lease):
            ipv4 = None
            ipv6 = None
            mac = lease["mac"]
            if vm.conn.is_test():
                # Hack it to match our interface for UI testing
                mac = iface.macaddr
            if mac == iface.macaddr:
                if lease["type"] == 0:
                    ipv4 = lease["ipaddr"]
                elif lease["type"] == 1:
                    ipv6 = lease["ipaddr"]
            return ipv4, ipv6

        for datalist in [list(qemuga.values()), leases, list(arp.values())]:
            ipv4 = None
            ipv6 = None
            for data in datalist:
                if "expirytime" in data:
                    tmpipv4, tmpipv6 = extract_lease(data)
                else:
                    tmpipv4, tmpipv6 = extract_dom(data)
                ipv4 = tmpipv4 or ipv4
                ipv6 = tmpipv6 or ipv6
            if ipv4 or ipv6:
                return ipv4, ipv6
        return None, None


class vmmInspectionApplication:
    def __init__(self):
        self.name = None
        self.display_name = None
        self.epoch = None
        self.version = None
        self.release = None
        self.summary = None
        self.description = None


class vmmInspectionData:
    def __init__(self):
        self.os_type = None
        self.distro = None
        self.major_version = None
        self.minor_version = None
        self.hostname = None
        self.product_name = None
        self.product_variant = None
        self.icon = None
        self.applications = None
        self.errorstr = None
        self.package_format = None


class vmmDomainSnapshot(vmmLibvirtObject):
    """
    Class wrapping a virDomainSnapshot object
    """

    def __init__(self, conn, backend):
        vmmLibvirtObject.__init__(self, conn, backend, backend.getName(), DomainSnapshot)

    ##########################
    # Required class methods #
    ##########################

    def _conn_tick_poll_param(self):
        return None  # pragma: no cover

    def class_name(self):
        return "snapshot"  # pragma: no cover

    def _XMLDesc(self, flags):
        return self._backend.getXMLDesc(flags=flags)

    def _get_backend_status(self):
        return self._STATUS_ACTIVE

    ###########
    # Actions #
    ###########

    def delete(self, force=True):
        ignore = force
        self._backend.delete()

    def _state_str_to_int(self):
        state = self.get_xmlobj().state
        statemap = {
            "nostate": libvirt.VIR_DOMAIN_NOSTATE,
            "running": libvirt.VIR_DOMAIN_RUNNING,
            "blocked": libvirt.VIR_DOMAIN_BLOCKED,
            "paused": libvirt.VIR_DOMAIN_PAUSED,
            "shutdown": libvirt.VIR_DOMAIN_SHUTDOWN,
            "shutoff": libvirt.VIR_DOMAIN_SHUTOFF,
            "crashed": libvirt.VIR_DOMAIN_CRASHED,
            "pmsuspended": getattr(libvirt, "VIR_DOMAIN_PMSUSPENDED", 7),
        }
        return statemap.get(state, libvirt.VIR_DOMAIN_SHUTOFF)

    def run_status(self):
        status = self._state_str_to_int()
        return LibvirtEnumMap.pretty_run_status(status, False)

    def run_status_icon_name(self):
        status = self._state_str_to_int()
        if status not in LibvirtEnumMap.VM_STATUS_ICONS:  # pragma: no cover
            log.debug("Unknown status %d, using NOSTATE", status)
            status = libvirt.VIR_DOMAIN_NOSTATE
        return LibvirtEnumMap.VM_STATUS_ICONS[status]

    def is_running(self):
        """
        Captured state is a running domain.
        """
        return self._state_str_to_int() in [libvirt.VIR_DOMAIN_RUNNING]

    def has_run_state(self):
        """
        Captured state contains run state in addition to disk state.
        """
        return self._state_str_to_int() in [libvirt.VIR_DOMAIN_RUNNING, libvirt.VIR_DOMAIN_PAUSED]

    def is_current(self):
        return self._backend.isCurrent()

    def is_external(self):
        if self.get_xmlobj().memory_type == "external":
            return True
        for disk in self.get_xmlobj().disks:
            if disk.snapshot == "external":
                return True
        return False


class _vmmDomainSetTimeThread(vmmGObject):
    """
    A separate thread handling time setting operations as not to block the main
    UI.
    """

    def __init__(self, domain):
        vmmGObject.__init__(self)
        self._domain = domain
        self._do_cancel = threading.Event()
        self._do_cancel.clear()
        self._thread = None
        self._maxwait = 30
        self._sleep = 0.5

    def start(self):
        """
        Start time setting thread if setting time is supported by the
        connection. Stop the old thread first. May block until the old thread
        terminates.
        """
        self.stop()

        # Only run the API for qemu and test drivers, they are the only ones
        # that support it. This will save spamming logs with error output.
        if not self._domain.conn.is_qemu() and not self._domain.conn.is_test():
            return  # pragma: no cover

        # For qemu, only run the API if the VM has the qemu guest agent in
        # the XML.
        if self._domain.conn.is_qemu() and not self._domain.has_agent():
            return

        log.debug("Starting time setting thread")
        self._thread = threading.Thread(name="settime thread", target=self._do_loop)
        self._thread.start()

    def stop(self):
        """
        Signal running thread to terminate and wait for it to do so.
        """
        if not self._thread:
            return

        log.debug("Stopping time setting thread")
        self._do_cancel.set()
        # thread may be in a loop waiting for an agent to come online or just
        # waiting for a set time operation to finish
        self._thread.join()
        self._thread = None
        self._do_cancel.clear()

    def _wait_for_agent(self):
        # Setting time of a qemu domain can only work if an agent is
        # defined and online. We only get here if one is defined. So wait
        # for it to come online now.
        waited = 0
        while waited < self._maxwait and not self._domain.agent_ready():
            if waited == 0:
                log.debug("Waiting for qemu guest agent to come online...")

            # sleep some time and potentially abort
            if self._do_cancel.wait(self._sleep):
                return

            waited += self._sleep

        if not self._domain.agent_ready():  # pragma: no cover
            log.debug("Giving up on qemu guest agent for time sync")
            return

    def _do_loop(self):
        """
        Run the domain's set time operation. Potentially wait for a guest agent
        to come online beforehand.
        """
        if self._domain.conn.is_qemu():
            self._wait_for_agent()
        self._domain.set_time()

    def _cleanup(self):
        self.stop()


class vmmDomain(vmmLibvirtObject):
    """
    Class wrapping virDomain libvirt objects. Is also extended to be
    backed by a virtinst.Guest object for new VM 'customize before install'
    """

    __gsignals__ = {
        "resources-sampled": (vmmLibvirtObject.RUN_FIRST, None, []),
        "inspection-changed": (vmmLibvirtObject.RUN_FIRST, None, []),
    }

    def __init__(self, conn, backend, key):
        vmmLibvirtObject.__init__(self, conn, backend, key, Guest)

        self.cloning = False

        self._install_abort = False
        self._id = None
        self._uuid = None
        self._has_managed_save = None
        self._snapshot_list = None
        self._autostart = None
        self._domain_caps = None
        self._status_reason = None
        self._ipfetcher = _IPFetcher()

        self.managedsave_supported = False
        self._domain_state_supported = False

        self.inspection = vmmInspectionData()
        self._set_time_thread = _vmmDomainSetTimeThread(self)

    def _cleanup(self):
        for snap in self._snapshot_list or []:
            snap.cleanup()
        self._snapshot_list = None
        self._set_time_thread.cleanup()
        self._set_time_thread = None
        vmmLibvirtObject._cleanup(self)

    def _init_libvirt_state(self):
        self.managedsave_supported = self.conn.support.domain_managed_save(self._backend)
        self._domain_state_supported = self.conn.support.domain_state(self._backend)

        # Determine available XML flags (older libvirt versions will error
        # out if passed SECURE_XML, INACTIVE_XML, etc)
        (self._inactive_xml_flags, self._active_xml_flags) = self.conn.get_dom_flags(self._backend)

        # Prime caches
        info = self._backend.info()
        self._refresh_status(newstatus=info[0])
        self.has_managed_save()
        self.snapshots_supported()

        if (
            self.get_name() == "Domain-0"
            and self.get_uuid() == "00000000-0000-0000-0000-000000000000"
        ):
            # We don't want virt-manager to track Domain-0 since it
            # doesn't work with our UI. Raising an error will ensures it
            # is denylisted.
            raise RuntimeError("Can't track Domain-0 as a vmmDomain")  # pragma: no cover

    ###########################
    # Misc API getter methods #
    ###########################

    def reports_stats(self):
        return True

    def _using_events(self):
        return self.conn.using_domain_events

    def get_id(self):
        if self._id is None:
            self._id = self._backend.ID()
        return self._id

    def status(self):
        return self._normalize_status(self._get_status())

    def status_reason(self):
        if self._status_reason is None:
            self._status_reason = 1
            if self._domain_state_supported:
                self._status_reason = self._backend.state()[1]
        return self._status_reason

    # If manual shutdown or destroy specified, make sure we don't continue
    # install process
    def get_install_abort(self):
        return bool(self._install_abort)

    def has_nvram(self):
        return bool(self.get_xmlobj().is_uefi() or self.get_xmlobj().os.nvram)

    def has_tpm_state(self):
        return any(tpm.type == "emulator" for tpm in self.get_xmlobj().devices.tpm)

    def is_persistent(self):
        return bool(self._backend.isPersistent())

    def has_shared_mem(self):
        """
        Return a value for 'Enable shared memory' UI, and an error if
        the value is not editable
        """
        is_shared = False
        err = None
        domcaps = self.get_domain_capabilities()

        if self.xmlobj.cpu.cells:
            err = _("Can not change shared memory setting when <numa> is configured.")
        elif (
            not domcaps.supports_filesystem_virtiofs() or not domcaps.supports_memorybacking_memfd()
        ):
            err = _("Libvirt may not be new enough to support memfd.")
        else:
            is_shared = self.xmlobj.memoryBacking.source_type == "memfd"

        return is_shared, err

    ##################
    # Support checks #
    ##################

    def supports_domain_job_info(self):
        if self.conn.is_test():
            # jobinfo isn't actually supported but this tests more code
            return True
        return self.conn.support.domain_job_info(self._backend)

    def snapshots_supported(self):
        if not self.conn.support.domain_list_snapshots(self._backend):
            return _("Libvirt connection does not support snapshots.")

    def get_domain_capabilities(self):
        if not self._domain_caps:
            self._domain_caps = self.get_xmlobj().lookup_domcaps()
        return self._domain_caps

    #############################
    # Internal XML handling API #
    #############################

    def _invalidate_xml(self):
        vmmLibvirtObject._invalidate_xml(self)
        self._id = None
        self._status_reason = None
        self._has_managed_save = None

    def _lookup_device_to_define(self, xmlobj, origdev, for_hotplug):
        if for_hotplug:
            return origdev

        dev = xmlobj.find_device(origdev)
        if dev:
            return dev

        # If we are removing multiple dev from an active VM, a double
        # attempt may result in a lookup failure. If device is present
        # in the active XML, assume all is good.
        if self.get_xmlobj().find_device(origdev):  # pragma: no cover
            log.debug("Device in active config but not inactive config.")
            return

        raise RuntimeError(  # pragma: no cover
            _("Could not find specified device in the inactive VM configuration: %s")
            % repr(origdev)
        )

    def _process_device_define(self, editdev, xmlobj, do_hotplug):
        if do_hotplug:
            self.hotplug(device=editdev)
        else:
            self._redefine_xmlobj(xmlobj)

    def _copy_nvram_file(self, new_name):
        """
        We need to do this copy magic because there is no Libvirt storage API
        to rename storage volume.
        """
        if not self.has_nvram():
            return None, None

        old_nvram_path = self.get_xmlobj().os.nvram
        if not old_nvram_path:  # pragma: no cover
            # Probably using firmware=efi which doesn't put nvram
            # path in the XML on older libvirt. Build the implied path
            old_nvram_path = os.path.join(
                self.conn.get_backend().get_libvirt_data_root_dir(),
                self.conn.get_backend().get_uri_driver(),
                "nvram",
                "%s_VARS.fd" % self.get_name(),
            )
            log.debug(
                "Guest is expected to use <nvram> but we didn't "
                "find one in the XML. Generated implied path=%s",
                old_nvram_path,
            )

        if not DeviceDisk.path_definitely_exists(
            self.conn.get_backend(), old_nvram_path
        ):  # pragma: no cover
            log.debug(
                "old_nvram_path=%s but it doesn't appear to exist. "
                "skipping rename nvram duplication",
                old_nvram_path,
            )
            return None, None

        from virtinst import Cloner

        old_nvram = DeviceDisk(self.conn.get_backend())
        old_nvram.set_source_path(old_nvram_path)
        ext = os.path.splitext(old_nvram_path)[1]

        nvram_dir = os.path.dirname(old_nvram.get_source_path())
        new_nvram_path = os.path.join(
            nvram_dir, "%s_VARS%s" % (os.path.basename(new_name), ext or ".fd")
        )

        new_nvram = Cloner.build_clone_disk(old_nvram, new_nvram_path, True, False)

        new_nvram.build_storage(None)
        return new_nvram, old_nvram

    ##############################
    # Persistent XML change APIs #
    ##############################

    def rename_domain(self, new_name):
        if new_name == self.get_name():
            return
        Guest.validate_name(self.conn.get_backend(), str(new_name))

        new_nvram, old_nvram = self._copy_nvram_file(new_name)

        try:
            self.define_name(new_name)
        except Exception as error:
            if new_nvram:
                try:
                    new_nvram.get_vol_object().delete(0)
                except Exception as warn:  # pragma: no cover
                    log.debug("rename failed and new nvram was not removed: '%s'", warn)
            raise error

        if not new_nvram:
            return

        try:
            old_nvram.get_vol_object().delete(0)
        except Exception as warn:  # pragma: no cover
            log.debug("old nvram file was not removed: '%s'", warn)

        self.define_overview(nvram=new_nvram.get_source_path())

    # Device Add/Remove
    def add_device(self, devobj):
        """
        Redefine guest with appended device XML 'devxml'
        """
        xmlobj = self._make_xmlobj_to_define()
        xmlobj.add_device(devobj)
        self._redefine_xmlobj(xmlobj)

    def remove_device(self, devobj):
        """
        Remove passed device from the inactive guest XML
        """
        xmlobj = self._make_xmlobj_to_define()
        editdev = self._lookup_device_to_define(xmlobj, devobj, False)
        if not editdev:
            return  # pragma: no cover

        xmlobj.remove_device(editdev)

        self._redefine_xmlobj(xmlobj)

    def replace_device_xml(self, devobj, newxml):
        """
        When device XML is editing from the XML editor window.
        """
        do_hotplug = False
        devclass = devobj.__class__
        newdev = devclass(devobj.conn, parsexml=newxml)

        xmlobj = self._make_xmlobj_to_define()
        editdev = self._lookup_device_to_define(xmlobj, devobj, do_hotplug)
        if not editdev:
            return  # pragma: no cover

        xmlobj.devices.replace_child(editdev, newdev)
        self._redefine_xmlobj(xmlobj)
        return editdev, newdev

    ##########################
    # non-device XML editing #
    ##########################

    def define_cpu(
        self,
        vcpus=_SENTINEL,
        model=_SENTINEL,
        secure=_SENTINEL,
        sockets=_SENTINEL,
        cores=_SENTINEL,
        threads=_SENTINEL,
        clear_topology=_SENTINEL,
    ):
        guest = self._make_xmlobj_to_define()

        if vcpus != _SENTINEL:
            guest.vcpus = int(vcpus)
            guest.vcpu_current = int(vcpus)

        if clear_topology is True:
            guest.cpu.topology.clear()
        elif sockets != _SENTINEL:
            guest.cpu.topology.sockets = sockets
            guest.cpu.topology.cores = cores
            guest.cpu.topology.threads = threads

        if secure != _SENTINEL or model != _SENTINEL:
            guest.cpu.secure = secure
            if model in guest.cpu.SPECIAL_MODES:
                guest.cpu.set_special_mode(guest, model)
            else:
                guest.cpu.set_model(guest, model)

        self._redefine_xmlobj(guest)

    def _edit_shared_mem(self, guest, mem_shared):
        source_type = _SENTINEL
        access_mode = _SENTINEL

        if mem_shared:
            source_type = "memfd"
            access_mode = "shared"
        else:
            source_type = None
            access_mode = None

        if source_type != _SENTINEL:
            guest.memoryBacking.source_type = source_type
        if access_mode != _SENTINEL:
            guest.memoryBacking.access_mode = access_mode

    def define_memory(self, memory=_SENTINEL, maxmem=_SENTINEL, mem_shared=_SENTINEL):
        guest = self._make_xmlobj_to_define()

        if memory != _SENTINEL:
            guest.currentMemory = int(memory)
        if maxmem != _SENTINEL:
            guest.memory = int(maxmem)
        if mem_shared != _SENTINEL:
            self._edit_shared_mem(guest, mem_shared)

        self._redefine_xmlobj(guest)

    def define_overview(
        self,
        machine=_SENTINEL,
        description=_SENTINEL,
        title=_SENTINEL,
        loader=_SENTINEL,
        nvram=_SENTINEL,
        firmware=_SENTINEL,
    ):
        guest = self._make_xmlobj_to_define()

        old_machine = None
        if machine != _SENTINEL:
            old_machine = guest.os.machine
            guest.os.machine = machine
            self._domain_caps = None
        if description != _SENTINEL:
            guest.description = description or None
        if title != _SENTINEL:
            guest.title = title or None

        if loader != _SENTINEL and firmware != _SENTINEL:
            guest.os.firmware = firmware
            if loader is None:
                guest.os.loader = None

                # But if switching to firmware=efi we may need to
                # preserve NVRAM paths, so skip clearing all the properties
                # and let libvirt do it for us.
                if firmware is None:
                    guest.disable_uefi()
            else:
                # Implies UEFI
                guest.set_uefi_path(loader)

        if nvram != _SENTINEL:
            guest.os.nvram = nvram

        if old_machine == "pc" and guest.os.machine == "q35":
            guest.add_q35_pcie_controllers()

        elif old_machine == "q35" and guest.os.machine == "pc":
            for dev in guest.devices.controller:
                if dev.model in ["pcie-root", "pcie-root-port"]:
                    guest.remove_device(dev)

        self._redefine_xmlobj(guest)

    def define_os(self, os_name=_SENTINEL):
        guest = self._make_xmlobj_to_define()

        if os_name != _SENTINEL:
            guest.set_os_name(os_name)

        self._redefine_xmlobj(guest)

    def define_boot(
        self,
        boot_order=_SENTINEL,
        boot_menu=_SENTINEL,
        kernel=_SENTINEL,
        initrd=_SENTINEL,
        dtb=_SENTINEL,
        kernel_args=_SENTINEL,
        init=_SENTINEL,
        initargs=_SENTINEL,
    ):

        guest = self._make_xmlobj_to_define()
        if boot_order != _SENTINEL:
            legacy = not self.can_use_device_boot_order()
            guest.set_boot_order(boot_order, legacy=legacy)

        if boot_menu != _SENTINEL:
            guest.os.bootmenu_enable = bool(boot_menu)
        if init != _SENTINEL:
            guest.os.init = init
            guest.os.set_initargs_string(initargs)

        if kernel != _SENTINEL:
            guest.os.kernel = kernel or None
        if initrd != _SENTINEL:
            guest.os.initrd = initrd or None
        if dtb != _SENTINEL:
            guest.os.dtb = dtb or None
        if kernel_args != _SENTINEL:
            guest.os.kernel_args = kernel_args or None

        self._redefine_xmlobj(guest)

    ######################
    # Device XML editing #
    ######################

    def define_disk(
        self,
        devobj,
        do_hotplug,
        path=_SENTINEL,
        readonly=_SENTINEL,
        shareable=_SENTINEL,
        removable=_SENTINEL,
        cache=_SENTINEL,
        discard=_SENTINEL,
        bus=_SENTINEL,
        serial=_SENTINEL,
    ):
        xmlobj = self._make_xmlobj_to_define()
        editdev = self._lookup_device_to_define(xmlobj, devobj, do_hotplug)
        if not editdev:
            return  # pragma: no cover

        validate = False
        if path != _SENTINEL:
            editdev.set_source_path(path)
            if not do_hotplug:
                editdev.sync_path_props()
                validate = True

        if readonly != _SENTINEL:
            editdev.read_only = readonly
        if shareable != _SENTINEL:
            editdev.shareable = shareable
        if removable != _SENTINEL:
            editdev.removable = removable

        if serial != _SENTINEL:
            editdev.serial = serial or None
        if cache != _SENTINEL:
            editdev.driver_cache = cache or None
        if discard != _SENTINEL:
            editdev.driver_discard = discard or None

        if bus != _SENTINEL:
            editdev.change_bus(self.xmlobj, bus)

        if validate:
            editdev.validate()
        self._process_device_define(editdev, xmlobj, do_hotplug)

    def define_network(
        self,
        devobj,
        do_hotplug,
        ntype=_SENTINEL,
        source=_SENTINEL,
        mode=_SENTINEL,
        model=_SENTINEL,
        macaddr=_SENTINEL,
        linkstate=_SENTINEL,
        portgroup=_SENTINEL,
    ):
        xmlobj = self._make_xmlobj_to_define()
        editdev = self._lookup_device_to_define(xmlobj, devobj, do_hotplug)
        if not editdev:
            return  # pragma: no cover

        if ntype != _SENTINEL:
            editdev.source = None

            editdev.type = ntype
            editdev.source = source
            editdev.source_mode = mode or None
            editdev.portgroup = portgroup or None

        if model != _SENTINEL:
            if editdev.model != model:
                editdev.address.clear()
            editdev.model = model

        if macaddr != _SENTINEL:
            editdev.macaddr = macaddr

        if linkstate != _SENTINEL:
            editdev.link_state = "up" if linkstate else "down"

        self._process_device_define(editdev, xmlobj, do_hotplug)

    def define_graphics(
        self,
        devobj,
        do_hotplug,
        listen=_SENTINEL,
        port=_SENTINEL,
        passwd=_SENTINEL,
        gtype=_SENTINEL,
        gl=_SENTINEL,
        rendernode=_SENTINEL,
    ):
        xmlobj = self._make_xmlobj_to_define()
        editdev = self._lookup_device_to_define(xmlobj, devobj, do_hotplug)
        if not editdev:
            return  # pragma: no cover

        if listen != _SENTINEL:
            editdev.listen = listen
        if port != _SENTINEL:
            editdev.port = port
        if passwd != _SENTINEL:
            editdev.passwd = passwd
        if gtype != _SENTINEL:
            editdev.type = gtype
        if gl != _SENTINEL:
            editdev.gl = gl
        if rendernode != _SENTINEL:
            editdev.rendernode = rendernode

        self._process_device_define(editdev, xmlobj, do_hotplug)

    def define_sound(self, devobj, do_hotplug, model=_SENTINEL):
        xmlobj = self._make_xmlobj_to_define()
        editdev = self._lookup_device_to_define(xmlobj, devobj, do_hotplug)
        if not editdev:
            return  # pragma: no cover

        if model != _SENTINEL:
            if editdev.model != model:
                editdev.address.clear()
            editdev.model = model

        self._process_device_define(editdev, xmlobj, do_hotplug)

    def define_video(self, devobj, do_hotplug, model=_SENTINEL, accel3d=_SENTINEL):
        xmlobj = self._make_xmlobj_to_define()
        editdev = self._lookup_device_to_define(xmlobj, devobj, do_hotplug)
        if not editdev:
            return  # pragma: no cover

        if model != _SENTINEL and model != editdev.model:
            editdev.model = model
            editdev.address.clear()

            # Clear out heads/ram values so they reset to default. If
            # we ever allow editing these values in the UI we should
            # drop this
            editdev.vram = None
            editdev.heads = None
            editdev.ram = None
            editdev.vgamem = None
            editdev.accel3d = None

        if accel3d != _SENTINEL:
            editdev.accel3d = accel3d

        self._process_device_define(editdev, xmlobj, do_hotplug)

    def define_watchdog(self, devobj, do_hotplug, model=_SENTINEL, action=_SENTINEL):
        xmlobj = self._make_xmlobj_to_define()
        editdev = self._lookup_device_to_define(xmlobj, devobj, do_hotplug)
        if not editdev:
            return  # pragma: no cover

        if model != _SENTINEL:
            if editdev.model != model:
                editdev.address.clear()
            editdev.model = model

        if action != _SENTINEL:
            editdev.action = action

        self._process_device_define(editdev, xmlobj, do_hotplug)

    def define_smartcard(self, devobj, do_hotplug, model=_SENTINEL):
        xmlobj = self._make_xmlobj_to_define()
        editdev = self._lookup_device_to_define(xmlobj, devobj, do_hotplug)
        if not editdev:
            return  # pragma: no cover

        if model != _SENTINEL:
            editdev.mode = model
            editdev.type = None
            editdev.type = editdev.default_type()

        self._process_device_define(editdev, xmlobj, do_hotplug)

    def define_controller(self, devobj, do_hotplug, model=_SENTINEL):
        xmlobj = self._make_xmlobj_to_define()
        editdev = self._lookup_device_to_define(xmlobj, devobj, do_hotplug)
        if not editdev:
            return  # pragma: no cover

        def _change_model():
            if editdev.type == "usb":
                ctrls = xmlobj.devices.controller
                ctrls = [x for x in ctrls if (x.type == DeviceController.TYPE_USB)]
                for dev in ctrls:
                    xmlobj.remove_device(dev)

                if model == "ich9-ehci1":
                    for dev in DeviceController.get_usb2_controllers(xmlobj.conn):
                        xmlobj.add_device(dev)
                elif model == "usb3":
                    dev = DeviceController.get_usb3_controller(xmlobj.conn, xmlobj)
                    xmlobj.add_device(dev)
                else:
                    dev = DeviceController(xmlobj.conn)
                    dev.type = "usb"
                    dev.model = model
                    xmlobj.add_device(dev)

            else:
                editdev.model = model
                editdev.address.clear()
                self.hotplug(device=editdev)

        if model != _SENTINEL:
            _change_model()

        self._process_device_define(editdev, xmlobj, do_hotplug)

    def define_filesystem(self, devobj, do_hotplug, newdev=_SENTINEL):
        xmlobj = self._make_xmlobj_to_define()
        editdev = self._lookup_device_to_define(xmlobj, devobj, do_hotplug)
        if not editdev:
            return  # pragma: no cover

        if newdev != _SENTINEL:
            # pylint: disable=maybe-no-member
            editdev.type = newdev.type
            editdev.accessmode = newdev.accessmode
            editdev.driver_type = newdev.driver_type
            editdev.driver_format = newdev.driver_format
            editdev.readonly = newdev.readonly
            editdev.source_units = newdev.source_units
            editdev.source = newdev.source
            editdev.target = newdev.target

        self._process_device_define(editdev, xmlobj, do_hotplug)

    def define_hostdev(self, devobj, do_hotplug, rom_bar=_SENTINEL, startup_policy=_SENTINEL):
        xmlobj = self._make_xmlobj_to_define()
        editdev = self._lookup_device_to_define(xmlobj, devobj, do_hotplug)
        if not editdev:
            return  # pragma: no cover

        if rom_bar != _SENTINEL:
            editdev.rom_bar = rom_bar

        if startup_policy != _SENTINEL:
            editdev.startup_policy = startup_policy

        self._process_device_define(editdev, xmlobj, do_hotplug)

    def define_tpm(self, devobj, do_hotplug, newdev=_SENTINEL):
        xmlobj = self._make_xmlobj_to_define()
        editdev = self._lookup_device_to_define(xmlobj, devobj, do_hotplug)
        if not editdev:
            return  # pragma: no cover

        if newdev != _SENTINEL:
            editdev.model = newdev.model
            editdev.type = newdev.type
            editdev.version = newdev.version
            editdev.device_path = newdev.device_path

        self._process_device_define(editdev, xmlobj, do_hotplug)

    def define_vsock(self, devobj, do_hotplug, auto_cid=_SENTINEL, cid=_SENTINEL):
        xmlobj = self._make_xmlobj_to_define()
        editdev = self._lookup_device_to_define(xmlobj, devobj, do_hotplug)
        if not editdev:
            return  # pragma: no cover

        if auto_cid != _SENTINEL:
            editdev.auto_cid = auto_cid
        if cid != _SENTINEL:
            editdev.cid = cid

        self._process_device_define(editdev, xmlobj, do_hotplug)

    ####################
    # Hotplug routines #
    ####################

    def attach_device(self, devobj):
        """
        Hotplug device to running guest
        """
        if not self.is_active():
            return

        devxml = devobj.get_xml()
        log.debug("attach_device with xml=\n%s", devxml)
        self._backend.attachDevice(devxml)

    def detach_device(self, devobj):
        """
        Hotunplug device from running guest
        """
        if not self.is_active():
            return

        devxml = devobj.get_xml()
        log.debug("detach_device with xml=\n%s", devxml)
        self._backend.detachDevice(devxml)

    def _update_device(self, devobj, flags=None):
        if flags is None:
            flags = getattr(libvirt, "VIR_DOMAIN_DEVICE_MODIFY_LIVE", 1)

        xml = devobj.get_xml()
        log.debug("update_device with xml=\n%s", xml)

        if self.config.CLITestOptions.test_update_device_fail:
            raise RuntimeError("fake update device failure")
        self._backend.updateDeviceFlags(xml, flags)

    def hotplug(
        self,
        memory=_SENTINEL,
        maxmem=_SENTINEL,
        description=_SENTINEL,
        title=_SENTINEL,
        device=_SENTINEL,
    ):
        if not self.is_active():
            return

        def _hotplug_memory(val):
            if val != self.xmlobj.currentMemory:
                self._backend.setMemory(val)

        def _hotplug_maxmem(val):
            if val != self.xmlobj.memory:
                self._backend.setMaxMemory(val)

        def _hotplug_metadata(val, mtype):
            flags = libvirt.VIR_DOMAIN_AFFECT_LIVE | libvirt.VIR_DOMAIN_AFFECT_CONFIG
            self._backend.setMetadata(mtype, val, None, None, flags)

        if memory != _SENTINEL:
            log.debug(
                "Hotplugging curmem=%s maxmem=%s for VM '%s'", memory, maxmem, self.get_name()
            )

            actual_cur = self.xmlobj.currentMemory
            if maxmem < actual_cur:
                # Set current first to avoid error
                _hotplug_memory(memory)
                _hotplug_maxmem(maxmem)
            else:
                _hotplug_maxmem(maxmem)
                _hotplug_memory(memory)

        if description != _SENTINEL:
            _hotplug_metadata(description, libvirt.VIR_DOMAIN_METADATA_DESCRIPTION)
        if title != _SENTINEL:
            _hotplug_metadata(title, libvirt.VIR_DOMAIN_METADATA_TITLE)

        if device != _SENTINEL:
            self._update_device(device)

    ########################
    # Libvirt API wrappers #
    ########################

    def _conn_tick_poll_param(self):
        return "pollvm"

    def class_name(self):
        return "domain"

    def _define(self, xml):
        self.conn.define_domain(xml)

    def _XMLDesc(self, flags):
        return self._backend.XMLDesc(flags)

    def _get_backend_status(self):
        return self._backend.info()[0]

    def get_autostart(self):
        if self._autostart is None:
            self._autostart = self._backend.autostart()
        return self._autostart

    def set_autostart(self, val):
        self._backend.setAutostart(val)

        # Recache value
        self._autostart = None
        self.get_autostart()

    def job_info(self):
        if self.conn.is_test():
            return testmock.fake_job_info()
        # It's tough to hit this via uitests because it depends
        # on the job lasting more than a second
        return self._backend.jobInfo()  # pragma: no cover

    def abort_job(self):
        self._backend.abortJob()

    def open_console(self, devname, stream, flags=0):
        return self._backend.openConsole(devname, stream, flags)

    def open_graphics_fd(self, idx):
        flags = 0
        return self._backend.openGraphicsFD(idx, flags)

    def list_snapshots(self):
        if self._snapshot_list is None:
            newlist = []
            for rawsnap in self._backend.listAllSnapshots():
                obj = vmmDomainSnapshot(self.conn, rawsnap)
                obj.init_libvirt_state()
                newlist.append(obj)
            self._snapshot_list = newlist
        return self._snapshot_list[:]

    def get_current_snapshot(self):
        if self._backend.hasCurrentSnapshot(0):
            rawsnap = self._backend.snapshotCurrent(0)
            obj = vmmDomainSnapshot(self.conn, rawsnap)
            obj.init_libvirt_state()
            return obj

        return None

    @vmmLibvirtObject.lifecycle_action
    def revert_to_snapshot(self, snap):
        # no use trying to set the guest time if is going to be switched off
        # after reverting to the snapshot
        will_be_running = snap.is_running()
        self._backend.revertToSnapshot(snap.get_backend())
        # looking at the domain state after revert will always come back as
        # paused, so look at the snapshot state instead
        if will_be_running:
            self._async_set_time()

    def create_snapshot(self, xml, redefine=False, diskOnly=False):
        flags = 0
        if redefine:
            flags = flags | libvirt.VIR_DOMAIN_SNAPSHOT_CREATE_REDEFINE
        else:
            if diskOnly:
                flags = flags | libvirt.VIR_DOMAIN_SNAPSHOT_CREATE_DISK_ONLY
            log.debug("Creating snapshot flags=%s xml=\n%s", flags, xml)
        obj = self._backend.snapshotCreateXML(xml, flags)
        log.debug("returned new snapshot XML:\n%s", obj.getXMLDesc(0))

    def _get_agent(self):
        """
        Return agent channel object if it is defined.
        """
        for dev in self.xmlobj.devices.channel:
            if dev.type == "unix" and dev.target_name == dev.CHANNEL_NAME_QEMUGA:
                return dev
        return None

    def has_agent(self):
        """
        Return True if domain has a guest agent defined.
        """
        return self._get_agent() is not None

    def agent_ready(self):
        """
        Return connected state of an agent.
        """
        dev = self._get_agent()
        if not dev:
            return False

        target_state = dev.target_state
        if self.conn.is_test():
            # test driver doesn't report 'connected' state so hack it here
            target_state = "connected"
        return target_state == "connected"

    def refresh_snapshots(self):
        self._snapshot_list = None

    def get_interface_addresses(self, iface, source):
        ret = {}
        log.debug("Calling interfaceAddresses source=%s", source)
        try:
            ret = self._backend.interfaceAddresses(source)
        except Exception as e:
            log.debug("interfaceAddresses failed: %s", str(e))
        if self.conn.is_test():
            ret = testmock.fake_interface_addresses(iface, source)
        return ret

    def get_ips(self, iface):
        return self._ipfetcher.get(self, iface)

    def refresh_ips(self, iface):
        return self._ipfetcher.refresh(self, iface)

    def set_time(self):
        """
        Try to set VM time to the current value. This is typically useful when
        clock wasn't running on the VM for some time (e.g. during suspension or
        migration), especially if the time delay exceeds NTP tolerance.
        It is not guaranteed that the time is actually set (it depends on guest
        environment, especially QEMU agent presence) or that the set time is
        very precise (NTP in the guest should take care of it if needed).

        Heavily based on
        https://github.com/openstack/nova/commit/414df1e56ea9df700756a1732125e06c5d97d792.
        """
        t = time.time()
        seconds = int(t)
        nseconds = int((t - seconds) * 10**9)
        try:
            self._backend.setTime(time={"seconds": seconds, "nseconds": nseconds})
            log.debug("Successfully set guest time")
        except Exception as e:  # pragma: no cover
            log.debug("Failed to set time: %s", e)

    def _async_set_time(self):
        """
        Asynchronously try to set guest time and maybe wait for a guest agent
        to come online using a separate thread.
        """
        self._set_time_thread.start()

    def _cancel_set_time(self):
        """
        Cancel a running guest time setting operation
        """
        self._set_time_thread.stop()

    ########################
    # XML Parsing routines #
    ########################

    def is_container(self):
        return self.get_xmlobj().os.is_container()

    def is_xenpv(self):
        return self.get_xmlobj().os.is_xenpv()

    def is_hvm(self):
        return self.get_xmlobj().os.is_hvm()

    def get_uuid(self):
        if self._uuid is None:
            self._uuid = self._backend.UUIDString()
        return self._uuid

    def get_abi_type(self):
        return self.get_xmlobj().os.os_type

    def get_hv_type(self):
        return self.get_xmlobj().type

    def get_pretty_hv_type(self):
        return self.conn.pretty_hv(self.get_abi_type(), self.get_hv_type())

    def get_arch(self):
        return self.get_xmlobj().os.arch

    def get_init(self):
        import shlex

        init = self.get_xmlobj().os.init
        initargs = " ".join([shlex.quote(i.val) for i in self.get_xmlobj().os.initargs])
        return init, initargs

    def get_emulator(self):
        return self.get_xmlobj().emulator

    def get_machtype(self):
        return self.get_xmlobj().os.machine

    def get_name_or_title(self):
        title = self.get_title()
        if title:
            return title
        return self.get_name()

    def get_title(self):
        return self.get_xmlobj().title

    def get_description(self):
        return self.get_xmlobj().description

    def get_boot_order(self):
        legacy = not self.can_use_device_boot_order()
        return self.xmlobj.get_boot_order(legacy=legacy)

    def get_boot_menu(self):
        guest = self.get_xmlobj()
        return bool(guest.os.bootmenu_enable)

    def get_boot_kernel_info(self):
        guest = self.get_xmlobj()
        return (guest.os.kernel, guest.os.initrd, guest.os.dtb, guest.os.kernel_args)

    def get_interface_devices_norefresh(self):
        xmlobj = self.get_xmlobj(refresh_if_nec=False)
        return xmlobj.devices.interface

    def get_disk_devices_norefresh(self):
        xmlobj = self.get_xmlobj(refresh_if_nec=False)
        return xmlobj.devices.disk

    def serial_is_console_dup(self, serial):
        return DeviceConsole.get_console_duplicate(self.xmlobj, serial)

    def can_use_device_boot_order(self):
        # Return 'True' if guest can use new style boot device ordering
        return self.conn.support.conn_device_boot_order()

    def get_bootable_devices(self):
        # redirdev can also be marked bootable, but it should be rarely
        # used and clutters the UI
        return self.xmlobj.get_bootable_devices(exclude_redirdev=True)

    ############################
    # Domain lifecycle methods #
    ############################

    # All these methods are usually run asynchronously from threads, so
    # let's be extra careful and have anything which might touch UI
    # or GObject.props invoked in an idle callback

    @vmmLibvirtObject.lifecycle_action
    def shutdown(self):
        self._cancel_set_time()
        self._install_abort = True
        self._backend.shutdown()

    @vmmLibvirtObject.lifecycle_action
    def reboot(self):
        self._cancel_set_time()
        self._install_abort = True
        self._backend.reboot(0)

    @vmmLibvirtObject.lifecycle_action
    def destroy(self):
        self._cancel_set_time()
        self._install_abort = True
        self._backend.destroy()

    @vmmLibvirtObject.lifecycle_action
    def reset(self):
        self._cancel_set_time()
        self._install_abort = True
        self._backend.reset(0)

    @vmmLibvirtObject.lifecycle_action
    def startup(self):
        has_managed = self.has_managed_save()
        if self.config.CLITestOptions.test_vm_run_fail or (
            has_managed and self.config.CLITestOptions.test_managed_save
        ):
            raise RuntimeError("fake error for managed save")

        self._backend.create()
        if has_managed:
            self._async_set_time()

    @vmmLibvirtObject.lifecycle_action
    def suspend(self):
        self._cancel_set_time()
        self._backend.suspend()

    @vmmLibvirtObject.lifecycle_action
    def delete(self, force=True):
        """
        @force: True if we are deleting domain, False if we are renaming domain

        If the domain is renamed we need to keep the nvram file.
        """
        flags = 0
        if force:
            flags |= getattr(libvirt, "VIR_DOMAIN_UNDEFINE_SNAPSHOTS_METADATA", 0)
            flags |= getattr(libvirt, "VIR_DOMAIN_UNDEFINE_MANAGED_SAVE", 0)
            if self.has_nvram():
                flags |= getattr(libvirt, "VIR_DOMAIN_UNDEFINE_NVRAM", 0)
        else:
            if self.has_nvram():
                flags |= getattr(libvirt, "VIR_DOMAIN_UNDEFINE_KEEP_NVRAM", 0)
            if self.has_tpm_state() and self.conn.support.domain_undefine_keep_tpm():
                flags |= getattr(libvirt, "VIR_DOMAIN_UNDEFINE_KEEP_TPM", 0)
        try:
            self._backend.undefineFlags(flags)
        except libvirt.libvirtError:
            log.exception("libvirt undefineFlags failed, falling back to old style")
            self._backend.undefine()

    @vmmLibvirtObject.lifecycle_action
    def resume(self):
        self._backend.resume()
        self._async_set_time()

    @vmmLibvirtObject.lifecycle_action
    def save(self, meter=None):
        self._cancel_set_time()
        self._install_abort = True

        if meter:
            start_job_progress_thread(self, meter, _("Saving domain to disk"))

        if self.config.CLITestOptions.test_managed_save:
            time.sleep(1.2)
        self._backend.managedSave(0)

    def has_managed_save(self):
        if not self.managedsave_supported:
            return False  # pragma: no cover

        if self._has_managed_save is None:
            try:
                self._has_managed_save = self._backend.hasManagedSaveImage(0)
            except Exception as e:  # pragma: no cover
                if self.conn.support.is_libvirt_error_no_domain(e):
                    return False
                raise

        return self._has_managed_save

    def remove_saved_image(self):
        if not self.has_managed_save():
            return  # pragma: no cover
        self._backend.managedSaveRemove(0)
        self._has_managed_save = None

    def migrate(
        self,
        destconn,
        dest_uri=None,
        tunnel=False,
        unsafe=False,
        temporary=False,
        xml=None,
        meter=None,
    ):
        self._cancel_set_time()
        self._install_abort = True

        flags = 0
        flags |= libvirt.VIR_MIGRATE_LIVE

        if not temporary:
            flags |= libvirt.VIR_MIGRATE_PERSIST_DEST
            flags |= libvirt.VIR_MIGRATE_UNDEFINE_SOURCE

        if tunnel:
            flags |= libvirt.VIR_MIGRATE_PEER2PEER
            flags |= libvirt.VIR_MIGRATE_TUNNELLED

        if unsafe:
            flags |= libvirt.VIR_MIGRATE_UNSAFE

        libvirt_destconn = destconn.get_backend().get_conn_for_api_arg()
        log.debug(
            "Migrating: conn=%s flags=%s uri=%s tunnel=%s unsafe=%s temporary=%s",
            destconn,
            flags,
            dest_uri,
            tunnel,
            unsafe,
            temporary,
        )

        if meter:
            start_job_progress_thread(self, meter, _("Migrating domain"))

        params = {}
        if dest_uri and not tunnel:
            params[libvirt.VIR_MIGRATE_PARAM_URI] = dest_uri
        if xml:
            params[libvirt.VIR_MIGRATE_PARAM_DEST_XML] = xml

        if self.conn.is_test() and "TESTSUITE-FAKE" in (dest_uri or ""):
            # If using the test driver and a special URI, fake successful
            # migration so we can test more of the migration wizard
            time.sleep(1.2)
            if not xml:
                xml = self.get_xml_to_define()
            destconn.define_domain(xml).create()
            self.delete()
        elif tunnel:
            self._backend.migrateToURI3(dest_uri, params, flags)
        else:
            self._backend.migrate3(libvirt_destconn, params, flags)

        # Don't schedule any conn update, migrate dialog handles it for us

    ###################
    # Stats accessors #
    ###################

    def _get_stats(self):
        return self.conn.statsmanager.get_vm_statslist(self)

    def stats_memory(self):
        return self._get_stats().get_record("curmem")

    def cpu_time(self):
        return self._get_stats().get_record("cpuTime")

    def host_cpu_time_percentage(self):
        return self._get_stats().get_record("cpuHostPercent")

    def guest_cpu_time_percentage(self):
        return self._get_stats().get_record("cpuGuestPercent")

    def network_rx_rate(self):
        return self._get_stats().get_record("netRxRate")

    def network_tx_rate(self):
        return self._get_stats().get_record("netTxRate")

    def disk_read_rate(self):
        return self._get_stats().get_record("diskRdRate")

    def disk_write_rate(self):
        return self._get_stats().get_record("diskWrRate")

    def network_traffic_rate(self):
        return self.network_tx_rate() + self.network_rx_rate()

    def network_traffic_max_rate(self):
        stats = self._get_stats()
        return max(stats.netRxMaxRate, stats.netTxMaxRate, 10.0)

    def disk_io_rate(self):
        return self.disk_read_rate() + self.disk_write_rate()

    def disk_io_max_rate(self):
        stats = self._get_stats()
        return max(stats.diskRdMaxRate, stats.diskWrMaxRate, 10.0)

    def host_cpu_time_vector(self, limit=None):
        return self._get_stats().get_vector("cpuHostPercent", limit)

    def guest_cpu_time_vector(self, limit=None):
        return self._get_stats().get_vector("cpuGuestPercent", limit)

    def stats_memory_vector(self, limit=None):
        return self._get_stats().get_vector("currMemPercent", limit)

    def network_traffic_vectors(self, limit=None, ceil=None):
        if ceil is None:
            ceil = self.network_traffic_max_rate()
        return self._get_stats().get_in_out_vector("netRxRate", "netTxRate", limit, ceil)

    def disk_io_vectors(self, limit=None, ceil=None):
        if ceil is None:
            ceil = self.disk_io_max_rate()
        return self._get_stats().get_in_out_vector("diskRdRate", "diskWrRate", limit, ceil)

    ###################
    # Status helpers ##
    ###################

    def _normalize_status(self, status):
        if status == libvirt.VIR_DOMAIN_NOSTATE:
            return libvirt.VIR_DOMAIN_RUNNING  # pragma: no cover
        elif status == libvirt.VIR_DOMAIN_BLOCKED:
            return libvirt.VIR_DOMAIN_RUNNING  # pragma: no cover
        return status

    def is_active(self):
        return not self.is_shutoff()

    def is_shutoff(self):
        return self.status() == libvirt.VIR_DOMAIN_SHUTOFF

    def is_crashed(self):
        return self.status() == libvirt.VIR_DOMAIN_CRASHED

    def is_stoppable(self):
        return self.status() in [
            libvirt.VIR_DOMAIN_RUNNING,
            libvirt.VIR_DOMAIN_PAUSED,
            libvirt.VIR_DOMAIN_CRASHED,
            libvirt.VIR_DOMAIN_PMSUSPENDED,
        ]

    def is_destroyable(self):
        return self.is_stoppable() or self.status() in [libvirt.VIR_DOMAIN_CRASHED]

    def is_runable(self):
        return self.is_shutoff()

    def is_pauseable(self):
        return self.status() in [libvirt.VIR_DOMAIN_RUNNING]

    def is_unpauseable(self):
        return self.status() in [libvirt.VIR_DOMAIN_PAUSED]

    def is_paused(self):
        return self.status() in [libvirt.VIR_DOMAIN_PAUSED]

    def is_cloneable(self):
        return self.status() in [libvirt.VIR_DOMAIN_SHUTOFF]

    def run_status(self):
        return LibvirtEnumMap.pretty_run_status(self.status(), self.has_managed_save())

    def run_status_reason(self):
        return LibvirtEnumMap.pretty_status_reason(self.status(), self.status_reason())

    def run_status_icon_name(self):
        status = self.status()
        if status not in LibvirtEnumMap.VM_STATUS_ICONS:  # pragma: no cover
            log.debug("Unknown status %s, using NOSTATE", status)
            status = libvirt.VIR_DOMAIN_NOSTATE
        return LibvirtEnumMap.VM_STATUS_ICONS[status]

    def set_inspection_data(self, data):
        self.inspection = data
        self.idle_emit("inspection-changed")

    ##################
    # config helpers #
    ##################

    def on_console_scaling_changed(self, *args, **kwargs):
        return self.config.listen_pervm(self.get_uuid(), "/scaling", *args, **kwargs)

    def set_console_scaling(self, value):
        self.config.set_pervm(self.get_uuid(), "/scaling", value)

    def get_console_scaling(self):
        ret = self.config.get_pervm(self.get_uuid(), "/scaling")
        if ret == -1:
            return self.config.get_console_scaling()
        return ret

    def on_console_resizeguest_changed(self, *args, **kwargs):
        return self.config.listen_pervm(self.get_uuid(), "/resize-guest", *args, **kwargs)

    def set_console_resizeguest(self, value):
        self.config.set_pervm(self.get_uuid(), "/resize-guest", value)

    def get_console_resizeguest(self):
        ret = self.config.get_pervm(self.get_uuid(), "/resize-guest")
        if ret == -1:
            return self.config.get_console_resizeguest()
        return ret

    def on_console_autoconnect_changed(self, *args, **kwargs):
        return self.config.listen_pervm(self.get_uuid(), "/resize-guest", *args, **kwargs)

    def set_console_autoconnect(self, value):
        self.config.set_pervm(self.get_uuid(), "/autoconnect", value)

    def get_console_autoconnect(self):
        ret = self.config.get_pervm(self.get_uuid(), "/autoconnect")
        if ret == -1:
            return self.config.get_console_autoconnect()
        return ret

    def set_details_window_size(self, w, h):
        self.config.set_pervm(self.get_uuid(), "/vm-window-size", (w, h))

    def get_details_window_size(self):
        ret = self.config.get_pervm(self.get_uuid(), "/vm-window-size")
        return ret

    def get_console_username(self):
        return self.config.get_pervm(self.get_uuid(), "/console-username")

    def set_console_username(self, username):
        return self.config.set_pervm(self.get_uuid(), "/console-username", username)

    def del_console_username(self):
        return self.config.set_pervm(self.get_uuid(), "/console-username", "")

    def get_cache_dir(self):
        ret = os.path.join(self.conn.get_cache_dir(), self.get_uuid())
        os.makedirs(ret, 0o755, exist_ok=True)
        return ret

    ###################
    # Polling helpers #
    ###################

    def tick(self, stats_update=True):
        if not self._using_events() and not stats_update:
            return

        dosignal = False
        if not self._using_events():
            # For domains it's pretty important that we are always using
            # the latest XML, but other objects probably don't want to do
            # this since it could be a performance hit.
            self._invalidate_xml()
            info = self._backend.info()
            dosignal = self._refresh_status(newstatus=info[0], cansignal=False)

        if stats_update:
            self.conn.statsmanager.refresh_vm_stats(self)
        if dosignal:
            self.idle_emit("state-changed")
        if stats_update:
            self.idle_emit("resources-sampled")


########################
# Libvirt domain class #
########################


class vmmDomainVirtinst(vmmDomain):
    """
    Domain object backed by a virtinst Guest object.

    Used for launching a details window for customizing a VM before install.
    """

    def __init__(self, conn, backend, key, installer):
        vmmDomain.__init__(self, conn, backend, key)
        self._orig_xml = None
        self._orig_backend = self._backend
        self._installer = installer

        self._refresh_status()
        log.debug("%s initialized with XML=\n%s", self, self._XMLDesc(0))

    def get_name(self):
        return self._backend.name

    def get_uuid(self):
        return self._backend.uuid

    def get_id(self):
        return -1  # pragma: no cover

    def has_managed_save(self):
        return False

    def snapshots_supported(self):
        return False

    def get_autostart(self):
        return self._installer.autostart

    def set_autostart(self, val):
        self._installer.autostart = bool(val)
        self.emit("state-changed")

    def _using_events(self):
        return False

    def _get_backend_status(self):
        return libvirt.VIR_DOMAIN_SHUTOFF

    def _cleanup(self):
        self._orig_backend = None
        self._installer = None
        super()._cleanup()

    ################
    # XML handling #
    ################

    def _sync_disk_storage_params(self, origdisk, newdisk):
        """
        When raw disk XML is edited from the customize wizard, the
        original DeviceDisk is completely blown away, but that will
        lose the storage creation info. This syncs that info across
        to the new DeviceDisk
        """
        if origdisk.get_source_path() != newdisk.get_source_path():
            return

        if origdisk.get_vol_object():
            log.debug(
                "Syncing vol_object=%s from origdisk=%s to newdisk=%s",
                origdisk.get_vol_object(),
                origdisk,
                newdisk,
            )
            newdisk.set_vol_object(origdisk.get_vol_object(), origdisk.get_parent_pool())
        elif origdisk.get_vol_install():
            log.debug(
                "Syncing vol_install=%s from origdisk=%s to newdisk=%s",
                origdisk.get_vol_install(),
                origdisk,
                newdisk,
            )
            newdisk.set_vol_install(origdisk.get_vol_install())

    def _replace_domain_xml(self, newxml):
        """
        Blow away the Guest instance we are tracking internally with
        a new one from the xmleditor UI, and sync over all disk storage
        info afterwards
        """
        newbackend = Guest(self._backend.conn, parsexml=newxml)

        for origdisk in self._backend.devices.disk:
            for newdisk in newbackend.devices.disk:
                if origdisk.compare_device(newdisk, newdisk.get_xml_idx()):
                    self._sync_disk_storage_params(origdisk, newdisk)
                    break

        self._backend = newbackend

    def replace_device_xml(self, devobj, newxml):
        """
        Overwrite vmmDomain's implementation, since we need to wire in
        syncing disk details.
        """
        if self._backend == self._orig_backend:
            # If the backend hasn't been replace yet, do it, so we don't
            # have a mix of is_build Guest with XML parsed objects which
            # might contain dragons
            self._replace_domain_xml(self._backend.get_xml())
        editdev, newdev = vmmDomain.replace_device_xml(self, devobj, newxml)
        if editdev.DEVICE_TYPE == "disk":
            self._sync_disk_storage_params(editdev, newdev)

    def define_xml(self, xml):
        origxml = self._backend.get_xml()
        self._replace_domain_xml(xml)
        self._redefine_xml_internal(origxml, xml)

    def define_name(self, newname):
        # We need to overwrite this, since the implementation for libvirt
        # needs to do some crazy stuff.
        xmlobj = self._make_xmlobj_to_define()
        xmlobj.name = str(newname)
        self._redefine_xmlobj(xmlobj)

    def _XMLDesc(self, flags):
        ignore = flags
        return self._backend.get_xml()

    def _define(self, xml):
        ignore = xml
        self.emit("state-changed")

    def _invalidate_xml(self):
        vmmDomain._invalidate_xml(self)
        self._orig_xml = None

    def _make_xmlobj_to_define(self):
        if not self._orig_xml:
            self._orig_xml = self._backend.get_xml()
        return self._backend

    def _redefine_xmlobj(self, xmlobj):
        self._redefine_xml_internal(self._orig_xml or "", xmlobj.get_xml())

    def rename_domain(self, new_name):
        Guest.validate_name(self._backend.conn, str(new_name))
        self.define_name(new_name)



#
# Copyright 2010, 2013 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from ..logger import log

from ..xmlbuilder import XMLBuilder, XMLProperty, XMLChildProperty
from .. import xmlutil


###################################
# Misc child nodes for CPU domain #
###################################


class _CPUTopology(XMLBuilder):
    """
    Class for generating XML for <cpu> child node <topology>.
    """

    XML_NAME = "topology"
    _XML_PROP_ORDER = ["sockets", "dies", "cores", "threads"]

    sockets = XMLProperty("./@sockets", is_int=True)
    dies = XMLProperty("./@dies", is_int=True)
    cores = XMLProperty("./@cores", is_int=True)
    threads = XMLProperty("./@threads", is_int=True)

    # While `dies` is optional and defaults to 1 if omitted,
    # `sockets`, `cores`, and `threads` are mandatory.
    def set_defaults_from_vcpus(self, vcpus):
        # The hierarchy is sockets > dies > cores > threads.
        #
        # In real world silicon though it is rare to have
        # high socket/die counts, but common to have huge
        # core counts.
        #
        # Some OS will even refuse to use sockets over a
        # a certain count.
        #
        # Thus we prefer to expose cores to the guest rather
        # than sockets as the default for missing fields
        if not self.cores:
            self.cores = vcpus // self.total_vcpus()

        if not self.sockets:
            self.sockets = vcpus // self.total_vcpus()

        if not self.dies:
            self.dies = 1

        if not self.threads:
            self.threads = vcpus // self.total_vcpus()

        if self.total_vcpus() != vcpus:
            raise ValueError(
                _(
                    "Total CPUs implied by topology "
                    "(sockets=%(sockets)d * dies=%(dies)d * cores=%(cores)d * threads=%(threads)d "
                    "== %(total)d) does not match vCPU count %(vcpus)d"
                )
                % {
                    "sockets": self.sockets,
                    "dies": self.dies,
                    "cores": self.cores,
                    "threads": self.threads,
                    "total": self.total_vcpus(),
                    "vcpus": vcpus,
                }
            )

        return

    def total_vcpus(self):
        """
        Determine the CPU count represented by topology
        """
        return (self.sockets or 1) * (self.dies or 1) * (self.cores or 1) * (self.threads or 1)


# Note: CPU cache is weird. The documentation implies that multiples instances
# can be declared, one for each cache level one wishes to define. However,
# libvirt doesn't accept more than one <cache> element, so it's implemented
# with `is_single=True` for now (see actual CPU Domain below).
class _CPUCache(XMLBuilder):
    """
    Class for generating XML for <cpu> child node <cache>.
    """

    XML_NAME = "cache"
    _XML_PROP_ORDER = ["level", "mode"]

    level = XMLProperty("./@level", is_int=True)
    mode = XMLProperty("./@mode")


class _CPUFeature(XMLBuilder):
    """
    Class for generating XML for <cpu> child nodes <feature>.
    """

    XML_NAME = "feature"
    _XML_PROP_ORDER = ["policy", "name"]

    name = XMLProperty("./@name")
    policy = XMLProperty("./@policy")


class _CPUMaxphysaddr(XMLBuilder):
    """
    Class for generating XML for <cpu> child node <maxphysaddr>.
    """

    XML_NAME = "maxphysaddr"
    _XML_PROP_ORDER = ["mode", "bits"]

    mode = XMLProperty("./@mode")
    bits = XMLProperty("./@bits", is_int=True)


##############
# NUMA cells #
##############


class _NUMACellSibling(XMLBuilder):
    """
    Class for generating XML for <cpu><numa><cell><distances> child nodes
    <sibling>, describing the distances to other NUMA cells.
    """

    XML_NAME = "sibling"
    _XML_PROP_ORDER = ["id", "value"]

    id = XMLProperty("./@id", is_int=True)
    value = XMLProperty("./@value", is_int=True)


class _NUMACellCache(XMLBuilder):
    """
    Class for generating XML for <cpu><numa><cell> child nodes <cache>,
    describing caches for NUMA cells.
    """

    XML_NAME = "cache"
    _XML_PROP_ORDER = [
        "level",
        "associativity",
        "policy",
        "size_value",
        "size_unit",
        "line_value",
        "line_unit",
    ]

    level = XMLProperty("./@level", is_int=True)
    associativity = XMLProperty("./@associativity")
    policy = XMLProperty("./@policy")

    size_value = XMLProperty("./size/@value", is_int=True)
    size_unit = XMLProperty("./size/@unit")
    line_value = XMLProperty("./line/@value", is_int=True)
    line_unit = XMLProperty("./line/@unit")


class _NUMACell(XMLBuilder):
    """
    Class for generating XML for <cpu><numa> child nodes <cell> XML, describing
    NUMA cells.
    """

    XML_NAME = "cell"
    _XML_PROP_ORDER = ["id", "cpus", "memory", "unit", "memAccess", "discard", "siblings", "caches"]

    id = XMLProperty("./@id", is_int=True)
    cpus = XMLProperty("./@cpus")
    memory = XMLProperty("./@memory", is_int=True)
    unit = XMLProperty("./@unit")
    memAccess = XMLProperty("./@memAccess")
    discard = XMLProperty("./@discard", is_yesno=True)

    siblings = XMLChildProperty(_NUMACellSibling, relative_xpath="./distances")
    caches = XMLChildProperty(_NUMACellCache)


#######################################
# Interconnections between NUMA cells #
#######################################


class _NUMALatency(XMLBuilder):
    """
    Class for generating XML for <cpu><numa><cell><interconnects> child nodes
    <latency>, describing latency between two NUMA memory nodes.
    """

    XML_NAME = "latency"
    _XML_PROP_ORDER = ["initiator", "target", "cache", "type", "value", "unit"]

    # Note: While libvirt happily accepts XML with a unit= property, it is
    # currently ignored on <latency> nodes.
    initiator = XMLProperty("./@initiator", is_int=True)
    target = XMLProperty("./@target", is_int=True)
    cache = XMLProperty("./@cache", is_int=True)
    type = XMLProperty("./@type")
    value = XMLProperty("./@value", is_int=True)
    unit = XMLProperty("./@unit")


class _NUMABandwidth(XMLBuilder):
    """
    Class for generating XML for <cpu><numa><cell><interconnects> child nodes
    <bandwidth>, describing bandwidth between two NUMA memory nodes.
    """

    XML_NAME = "bandwidth"
    _XML_PROP_ORDER = ["initiator", "target", "cache", "type", "value", "unit"]

    # Note: The documentation only mentions <latency> nodes having a cache=
    # attribute, but <bandwidth> and <latency> nodes are otherwise identical
    # and libvirt will happily accept XML with a cache= attribute on
    # <bandwidth> nodes as well, so let's leave it here for now.
    initiator = XMLProperty("./@initiator", is_int=True)
    target = XMLProperty("./@target", is_int=True)
    cache = XMLProperty("./@cache", is_int=True)
    type = XMLProperty("./@type")
    value = XMLProperty("./@value", is_int=True)
    unit = XMLProperty("./@unit")


#####################
# Actual CPU domain #
#####################


class DomainCpu(XMLBuilder):
    """
    Class for generating <cpu> XML
    """

    XML_NAME = "cpu"
    _XML_PROP_ORDER = [
        "mode",
        "match",
        "check",
        "migratable",
        "model",
        "model_fallback",
        "model_vendor_id",
        "vendor",
        "topology",
        "cache",
        "features",
        "cells",
        "latencies",
        "bandwidths",
        "maxphysaddr",
    ]

    ##################
    # XML properties #
    ##################

    # Note: This is not a libvirt property. This is specific to the virt-*
    # tools and causes additional security features to be added to the VM.
    # See the security mitigation related functions below for more details.
    secure = True

    mode = XMLProperty("./@mode")
    match = XMLProperty("./@match")
    check = XMLProperty("./@check")
    migratable = XMLProperty("./@migratable", is_onoff=True)

    model = XMLProperty("./model")
    model_fallback = XMLProperty("./model/@fallback")
    model_vendor_id = XMLProperty("./model/@vendor_id")
    vendor = XMLProperty("./vendor")

    topology = XMLChildProperty(_CPUTopology, is_single=True)
    cache = XMLChildProperty(_CPUCache, is_single=True)
    features = XMLChildProperty(_CPUFeature)

    # NUMA related properties
    cells = XMLChildProperty(_NUMACell, relative_xpath="./numa")
    latencies = XMLChildProperty(_NUMALatency, relative_xpath="./numa/interconnects")
    bandwidths = XMLChildProperty(_NUMABandwidth, relative_xpath="./numa/interconnects")

    maxphysaddr = XMLChildProperty(_CPUMaxphysaddr, is_single=True)

    #############################
    # Special CPU mode handling #
    #############################

    special_mode_was_set = False
    # These values are exposed on the command line, so are stable API
    SPECIAL_MODE_HOST_MODEL_ONLY = "host-model-only"
    SPECIAL_MODE_HV_DEFAULT = "hv-default"
    SPECIAL_MODE_HOST_COPY = "host-copy"
    SPECIAL_MODE_HOST_MODEL = "host-model"
    SPECIAL_MODE_HOST_PASSTHROUGH = "host-passthrough"
    SPECIAL_MODE_MAXIMUM = "maximum"
    SPECIAL_MODE_CLEAR = "clear"
    SPECIAL_MODE_APP_DEFAULT = "default"
    SPECIAL_MODES = [
        SPECIAL_MODE_HOST_MODEL_ONLY,
        SPECIAL_MODE_HV_DEFAULT,
        SPECIAL_MODE_HOST_COPY,
        SPECIAL_MODE_HOST_MODEL,
        SPECIAL_MODE_HOST_PASSTHROUGH,
        SPECIAL_MODE_MAXIMUM,
        SPECIAL_MODE_CLEAR,
        SPECIAL_MODE_APP_DEFAULT,
    ]

    def _should_use_maximum_cpu_mode(self, guest, domcaps):
        if (
            domcaps.supports_maximum_cpu_mode()
            and guest.type == "qemu"
            and (
                guest.os.is_x86()
                or guest.os.is_arm_machvirt()
                or guest.os.is_riscv_virt()
                or guest.os.is_loongarch64()
            )
        ):
            return True

        return False

    def _get_app_default_mode(self, guest):
        # Depending on if libvirt+qemu is new enough, we prefer
        # host-passthrough, then host-model, and finally host-model-only
        # Emulated guests use maximum mode if available
        domcaps = guest.lookup_domcaps()

        if self._should_use_maximum_cpu_mode(guest, domcaps):
            return self.SPECIAL_MODE_MAXIMUM

        if domcaps.supports_safe_host_passthrough():
            return self.SPECIAL_MODE_HOST_PASSTHROUGH

        log.debug("Safe host-passthrough is not available")
        if domcaps.supports_safe_host_model():
            return self.SPECIAL_MODE_HOST_MODEL

        log.debug("Safe host-model is not available")
        return self.SPECIAL_MODE_HOST_MODEL_ONLY

    def set_special_mode(self, guest, val):
        if val == self.SPECIAL_MODE_APP_DEFAULT:
            val = self._get_app_default_mode(guest)
            log.debug("Using default cpu mode=%s", val)

        if (
            val == self.SPECIAL_MODE_HOST_MODEL
            or val == self.SPECIAL_MODE_HOST_PASSTHROUGH
            or val == self.SPECIAL_MODE_MAXIMUM
        ):
            self.model = None
            self.vendor = None
            self.model_fallback = None
            self.migratable = None
            self.check = None
            for f in self.features:
                self.remove_child(f)
            self.mode = val
        elif val == self.SPECIAL_MODE_HV_DEFAULT or val == self.SPECIAL_MODE_CLEAR:
            self.clear()
        elif val == self.SPECIAL_MODE_HOST_MODEL_ONLY or val == self.SPECIAL_MODE_HOST_COPY:
            if val == self.SPECIAL_MODE_HOST_COPY:
                log.warning(
                    "CPU mode=%s no longer supported, using mode=%s",
                    val,
                    self.SPECIAL_MODE_HOST_MODEL_ONLY,
                )
            if self.conn.caps.host.cpu.model:
                self.clear()
                self.set_model(guest, self.conn.caps.host.cpu.model)
        else:
            raise xmlutil.DevError("unknown special cpu mode '%s'" % val)

        self.special_mode_was_set = True

    ########################
    # Security mitigations #
    ########################

    def _add_security_features(self, guest):
        domcaps = guest.lookup_domcaps()
        for feature in domcaps.get_cpu_security_features():
            exists = False
            for f in self.features:
                if f.name == feature:
                    exists = True
                    break
            if not exists:
                self.add_feature(feature)

    def check_security_features(self, guest):
        """
        Since 'secure' property is not exported into the domain XML
        we might need to refresh its state.
        """
        domcaps = guest.lookup_domcaps()
        features = domcaps.get_cpu_security_features()
        if len(features) == 0:
            self.secure = False
            return

        guestFeatures = [f.name for f in self.features if f.policy == "require"]
        if self.model:
            if self.model.endswith("IBRS"):
                guestFeatures.append("spec-ctrl")
            if self.model.endswith("IBPB"):
                guestFeatures.append("ibpb")

        self.secure = set(features) <= set(guestFeatures)

    def _remove_security_features(self, guest):
        domcaps = guest.lookup_domcaps()
        for feature in domcaps.get_cpu_security_features():
            for f in self.features:
                if f.name == feature and f.policy == "require":
                    self.remove_child(f)
                    break

    ###########
    # Helpers #
    ###########

    def set_model(self, guest, val):
        log.debug("setting cpu model %s", val)
        self.migratable = None
        if val:
            self.mode = "custom"
            if not self.match:
                self.match = "exact"
            if self.secure:
                self._add_security_features(guest)
            else:
                self._remove_security_features(guest)
        self.model = val

    def add_feature(self, name, policy="require"):
        feature = self.features.add_new()
        feature.name = name
        feature.policy = policy

    def vcpus_from_topology(self):
        """
        Determine the CPU count represented by topology, or 1 if
        no topology is set
        """
        return self.topology.total_vcpus()

    def has_topology(self):
        """
        Return True if any topology info is set
        """
        return bool(self.topology.get_xml())

    def set_topology_defaults(self, vcpus, create=False):
        """
        Fill in unset topology values, using the passed vcpus count.
        If @create is False, this will not set topology from scratch,
        just fill in missing topology values.
        If @create is True, this will create topology from scratch.
        """
        if not self.has_topology() and not create:
            return
        self.topology.set_defaults_from_vcpus(vcpus)

    ##################
    # Default config #
    ##################

    def _validate_default_host_model_only(self, guest):
        # It's possible that the value HOST_MODEL_ONLY gets from
        # <capabilities> is not actually supported by qemu/kvm
        # combo which will be reported in <domainCapabilities>
        if not self.model:
            return  # pragma: no cover

        domcaps = guest.lookup_domcaps()
        domcaps_mode = domcaps.cpu.get_mode("custom")
        if not domcaps_mode:
            return  # pragma: no cover

        cpu_model = domcaps_mode.get_model(self.model)
        if cpu_model and cpu_model.usable != "no":
            return

        log.debug(
            "Host capabilities CPU '%s' is not supported "
            "according to domain capabilities. Unsetting CPU model",
            self.model,
        )
        self.model = None

    def _set_cpu_x86_kvm_default(self, guest):
        if guest.os.arch != self.conn.caps.host.cpu.arch:
            return

        mode = guest.x86_cpu_default

        self.set_special_mode(guest, mode)
        if mode == self.SPECIAL_MODE_HOST_MODEL_ONLY:
            self._validate_default_host_model_only(guest)

    def set_defaults(self, guest):
        if not self.conn.is_test() and not self.conn.is_qemu():
            return
        if self.special_mode_was_set:
            return
        if self.model or self.mode:
            return

        if guest.os.is_arm_machvirt() and guest.type == "kvm":
            self.mode = self.SPECIAL_MODE_HOST_PASSTHROUGH

        elif guest.os.is_x86() and guest.type == "kvm":
            self._set_cpu_x86_kvm_default(guest)

        else:
            domcaps = guest.lookup_domcaps()

            # Prefer to emulate a feature-rich CPU instead of a basic one
            if self._should_use_maximum_cpu_mode(guest, domcaps):
                self.set_special_mode(guest, self.SPECIAL_MODE_MAXIMUM)



# Copyright(c) FUJITSU Limited 2007.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import argparse
import sys

from . import cli
from .cli import fail, print_stdout, print_stderr
from .cloner import Cloner


def _process_src(options):
    src_name = options.src_name
    src_xml = None
    if options.original_xml:
        src_xml = open(options.original_xml).read()
    elif not src_name:
        fail(_("An original machine name is required, use '--original src_name' and try again."))
    return src_name, src_xml


def _process_macs(options, cloner):
    new_macs = options.new_mac
    if not new_macs or new_macs[0] == "RANDOM":
        return

    for mac in new_macs:
        cli.validate_mac(cloner.conn, mac)

    for iface in cloner.new_guest.devices.interface[:]:
        iface.macaddr = new_macs.pop(0)


def _process_disks(options, cloner):
    newpaths = (options.new_diskfile or [])[:]

    diskinfos = cloner.get_nonshare_diskinfos()
    for diskinfo in diskinfos:
        origpath = diskinfo.disk.get_source_path()
        newpath = None
        if newpaths:
            newpath = newpaths.pop(0)
        elif options.auto_clone:
            break

        if origpath is None:
            newpath = None
        diskinfo.set_new_path(newpath, options.sparse)
        diskinfo.raise_error()


def _validate_disks(cloner):
    # Extra CLI validation for specified disks
    for diskinfo in cloner.get_diskinfos():
        diskinfo.raise_error()
        if not diskinfo.new_disk:
            continue
        warn_overwrite = not diskinfo.is_preserve_requested()
        cli.validate_disk(diskinfo.new_disk, warn_overwrite=warn_overwrite)


def parse_args():
    desc = _(
        "Duplicate a virtual machine, changing all the unique "
        "host side configuration like MAC address, name, etc. \n\n"
        "The VM contents are NOT altered: virt-clone does not change "
        "anything _inside_ the guest OS, it only duplicates disks and "
        "does host side changes. So things like changing passwords, "
        "changing static IP address, etc are outside the scope of "
        "this tool. For these types of changes, please see virt-sysprep(1)."
    )
    parser = cli.setupParser("%(prog)s --original [NAME] ...", desc)
    cli.add_connect_option(parser)

    geng = parser.add_argument_group(_("General Options"))
    geng.add_argument(
        "-o", "--original", dest="src_name", help=_("Name of the original guest to clone.")
    )
    geng.add_argument("--original-xml", help=_("XML file to use as the original guest."))
    geng.add_argument(
        "--auto-clone",
        action="store_true",
        help=_("Auto generate clone name and storage paths from the original guest configuration."),
    )
    geng.add_argument("-n", "--name", dest="new_name", help=_("Name for the new guest"))
    geng.add_argument("-u", "--uuid", dest="new_uuid", help=argparse.SUPPRESS)
    geng.add_argument("--reflink", action="store_true", help=_("use btrfs COW lightweight copy"))

    stog = parser.add_argument_group(_("Storage Configuration"))
    stog.add_argument(
        "-f",
        "--file",
        dest="new_diskfile",
        action="append",
        help=_("New file to use as the disk image for the new guest"),
    )
    stog.add_argument(
        "--force-copy",
        dest="target",
        action="append",
        help=_("Force to copy devices (eg, if 'hdc' is a readonly cdrom device, --force-copy=hdc)"),
    )
    stog.add_argument(
        "--skip-copy",
        action="append",
        help=_(
            "Skip copy of the device target. (eg, if 'vda' is a "
            "disk you don't want to copy and use the same path "
            "in the new VM, use --skip-copy=vda)"
        ),
    )
    stog.add_argument(
        "--nonsparse",
        action="store_false",
        dest="sparse",
        default=True,
        help=_("Do not use a sparse file for the clone's disk image"),
    )
    stog.add_argument(
        "--preserve-data",
        dest="preserve",
        action="store_true",
        default=False,
        help=_(
            "Do not clone storage contents to specified file paths, "
            "their contents will be left untouched. "
            "This requires specifying existing paths for "
            "every cloneable disk image."
        ),
    )
    stog.add_argument(
        "--nvram", dest="new_nvram", help=_("New file to use as storage for nvram VARS")
    )

    netg = parser.add_argument_group(_("Networking Configuration"))
    netg.add_argument(
        "-m",
        "--mac",
        dest="new_mac",
        action="append",
        help=_("New fixed MAC address for the clone guest. Default is a randomly generated MAC"),
    )

    misc = parser.add_argument_group(_("Miscellaneous Options"))

    # Just used for clone tests
    misc.add_argument(
        "--__test-nodry",
        action="store_true",
        dest="test_nodry",
        default=False,
        help=argparse.SUPPRESS,
    )

    cli.add_misc_options(misc, prompt=True, replace=True, printxml=True)

    cli.autocomplete(parser)

    return parser.parse_args()


def main(conn=None):
    cli.earlyLogging()
    options = parse_args()

    options.quiet = options.quiet or options.xmlonly
    cli.setupLogging("virt-clone", options.debug, options.quiet)

    cli.convert_old_force(options)
    cli.parse_check(options.check)
    cli.set_prompt(options.prompt)
    conn = cli.getConnection(options.connect, conn=conn)

    if options.new_diskfile is None and options.auto_clone is False:
        fail(
            _(
                "Either --auto-clone or --file is required,"
                " use '--auto-clone or --file' and try again."
            )
        )

    src_name, src_xml = _process_src(options)
    cloner = Cloner(conn, src_name, src_xml)

    cloner.set_replace(bool(options.replace))
    cloner.set_reflink(bool(options.reflink))
    cloner.set_sparse(bool(options.sparse))

    if options.new_uuid is not None:
        cloner.set_clone_uuid(options.new_uuid)
    if options.new_nvram:
        cloner.set_nvram_path(options.new_nvram)

    force_targets = options.target or []
    skip_targets = options.skip_copy or []
    for diskinfo in cloner.get_diskinfos():
        if diskinfo.disk.target in force_targets:
            diskinfo.set_clone_requested()
        if diskinfo.disk.target in skip_targets:
            diskinfo.set_share_requested()

    if options.preserve:
        for diskinfo in cloner.get_nonshare_diskinfos():
            diskinfo.set_preserve_requested()
        if cloner.nvram_diskinfo:
            cloner.nvram_diskinfo.set_preserve_requested()

    if options.new_name:
        cloner.set_clone_name(options.new_name)
    elif not options.auto_clone:
        fail(
            _(
                "A name is required for the new virtual machine,"
                " use '--name NEW_VM_NAME' to specify one."
            )
        )

    _process_macs(options, cloner)
    _process_disks(options, cloner)

    cloner.prepare()

    _validate_disks(cloner)

    run = True
    if options.xmlonly:
        run = options.test_nodry
        print_stdout(cloner.new_guest.get_xml(), do_force=True)
    if run:
        cloner.start_duplicate(cli.get_meter())
        print_stdout("")
        print_stdout(_("Clone '%s' created successfully.") % cloner.new_guest.name)

    return 0


def runcli():  # pragma: no cover
    try:
        sys.exit(main())
    except SystemExit as sys_e:
        sys.exit(sys_e.code)
    except KeyboardInterrupt:
        print_stderr(_("Installation aborted at user request"))
    except Exception as main_e:
        fail(main_e)



#
# Copyright 2006-2009, 2013, 2014 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os

from .. import progress
from ..devices import DeviceDisk
from ..logger import log
from ..storage import StoragePool, StorageVolume


def _build_pool(conn, meter, path):
    """
    Helper function for building a pool on demand. Used for building
    a scratchdir pool for volume upload
    """
    pool = StoragePool.lookup_pool_by_path(conn, path)
    if pool:  # pragma: no cover
        log.debug("Existing pool '%s' found for %s", pool.name(), path)
        StoragePool.ensure_pool_is_running(pool, refresh=True)
        return pool

    name = StoragePool.find_free_name(conn, "boot-scratch")
    log.debug("Building storage pool: path=%s name=%s", path, name)
    poolbuild = StoragePool(conn)
    poolbuild.type = poolbuild.TYPE_DIR
    poolbuild.name = name
    poolbuild.target_path = path

    # Explicitly don't build? since if we are creating this directory
    # we probably don't have correct perms
    ret = poolbuild.install(meter=meter, create=True, build=False, autostart=True)
    return ret


class _MockStream:
    _data_size = None

    def send(self, data):
        if self._data_size is None:
            self._data_size = len(data)

        block_size = 128
        ret = min(self._data_size, block_size)
        self._data_size = max(0, self._data_size - block_size)
        return ret

    def finish(self):
        pass


def _upload_file(conn, meter, destpool, src):
    """
    Helper for uploading a file to a pool, via libvirt. Used for
    kernel/initrd upload when we can't access the system scratchdir
    """
    # Build stream object
    if conn.in_testsuite():
        stream = _MockStream()
    else:
        stream = conn.newStream(0)  # pragma: no cover

    def safe_send(data):
        while True:
            ret = stream.send(data)
            if ret == 0 or ret == len(data):
                break
            data = data[ret:]

    meter = progress.ensure_meter(meter)

    # Build placeholder volume
    size = os.path.getsize(src)
    basename = os.path.basename(src)
    name = StorageVolume.find_free_name(conn, destpool, basename)
    log.debug("Generated volume name %s", name)

    vol_install = DeviceDisk.build_vol_install(
        conn, name, destpool, (float(size) / 1024.0 / 1024.0 / 1024.0), True
    )

    disk = DeviceDisk(conn)
    disk.set_vol_install(vol_install)
    disk.validate()

    disk.build_storage(meter)
    vol = disk.get_vol_object()
    if not vol:
        raise RuntimeError("Failed to lookup scratch media volume")  # pragma: no cover

    try:
        # Register upload
        offset = 0
        length = size
        flags = 0
        if not conn.in_testsuite():
            vol.upload(stream, offset, length, flags)  # pragma: no cover

        # Open source file
        fileobj = open(src, "rb")

        # Start transfer
        total = 0
        msg = _("Transferring '%(filename)s'") % {"filename": os.path.basename(src)}
        meter.start(msg, size)
        while True:
            blocksize = 1024 * 1024  # 1 MiB
            data = fileobj.read(blocksize)
            if not data:
                break

            safe_send(data)
            total += len(data)
            meter.update(total)

        # Cleanup
        stream.finish()
        meter.end()
    except Exception:  # pragma: no cover
        vol.delete(0)
        raise

    return vol


def upload_paths(conn, system_scratchdir, meter, pathlist):
    """
    Upload passed paths to the connection scratchdir
    """
    # Build pool
    log.debug("Uploading kernel/initrd media")
    pool = _build_pool(conn, meter, system_scratchdir)

    tmpvols = []
    newpaths = []
    try:
        for path in pathlist:
            vol = _upload_file(conn, meter, pool, path)
            newpaths.append(vol.path())
            tmpvols.append(vol)
    except Exception:  # pragma: no cover
        for vol in tmpvols:
            vol.delete(0)
        raise

    return newpaths, tmpvols



#
# Copyright 2006, 2013 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.
#

import os


class DevError(RuntimeError):
    def __init__(self, msg):
        RuntimeError.__init__(self, "programming error: %s" % msg)


def listify(lst):
    if lst is None:
        return []
    elif not isinstance(lst, list):
        return [lst]
    else:
        return lst


def xml_escape(xml):
    """
    Replaces chars ' " < > & with xml safe counterparts
    """
    if xml:
        xml = xml.replace("&", "&amp;")
        xml = xml.replace("'", "&apos;")
        xml = xml.replace('"', "&quot;")
        xml = xml.replace("<", "&lt;")
        xml = xml.replace(">", "&gt;")
    return xml


def get_prop_path(obj, prop_path):
    """Return value of attribute identified by `prop_path`

    Look up the attribute of `obj` identified by `prop_path`
    (separated by "."). If any component along the path is missing an
    `AttributeError` is raised.

    """
    parent = obj
    pieces = prop_path.split(".")
    for piece in pieces[:-1]:
        parent = getattr(parent, piece)

    return getattr(parent, pieces[-1])


def set_prop_path(obj, prop_path, value):
    """Set value of attribute identified by `prop_path`

    Set the attribute of `obj` identified by `prop_path` (separated by
    ".") to `value`. If any component along the path is missing an
    `AttributeError` is raised.
    """
    parent = obj
    pieces = prop_path.split(".")
    for piece in pieces[:-1]:
        parent = getattr(parent, piece)

    return setattr(parent, pieces[-1], value)


def in_testsuite():
    return "VIRTINST_TEST_SUITE" in os.environ


def diff(origstr, newstr, fromfile="Original", tofile="New"):
    import difflib

    dlist = difflib.unified_diff(
        origstr.splitlines(1), newstr.splitlines(1), fromfile=fromfile, tofile=tofile
    )
    return "".join(dlist)


def unindent_device_xml(xml):
    import re

    lines = xml.splitlines()
    if not lines:
        return xml  # pragma: no cover

    ret = ""
    unindent = 0
    for c in lines[0]:
        if c != " ":
            break
        unindent += 1

    for line in lines:
        if re.match(r"^%s *<.*$" % (unindent * " "), line):
            line = line[unindent:]
        ret += line + "\n"
    return ret
